#ifndef __FBGFX_ADDON_BI__
#define __FBGFX_ADDON_BI__

#include once "crt.bi"
#include once "fbgfx.bi"


#if __FB_LANG__ <> "fb"
#error 666: wrong FreeBASIC dialect !
#else
using FB
#endif

#ifdef __FB_WIN32__
# libpath "lib/win"
#else
# libpath "lib/lin"
#endif

#ifndef __FB_64BIT__
# inclib "FBImage-32-static"
#else
# inclib "FBImage-64-static"
#endif


extern "C"
' Load BMP, PNG, JPG, TGA, DDS from file or memory as FBImage
declare function LoadRGBAFile(byval filename as const zstring ptr) as any ptr
declare function LoadRGBAMemory(byval buffer as const any ptr, byval buffersize as long) as any ptr
declare function GetLastResult() as const zstring ptr

end extern



#ifndef PI
const as single PI = atn(1)*4
#endif

const as single RAD2DEG = 180/PI
const as single DEG2RAD = PI/180


union RGBAColor
  declare constructor(byval col as ulong=&HFF000000)
  declare constructor(byval r8 as ubyte, byval g8 as ubyte, byval b8 as ubyte, byval a8 as ubyte=255)
  declare function u8  as ubyte
  declare function u16 as ushort
  type field=1 
   as ubyte b,g,r,a
  end type
  as ulong u32
end union
constructor RGBAColor(byval col as ulong=&HFF000000)
  this.u32=col
end constructor

constructor RGBAColor(byval r8 as ubyte, byval g8 as ubyte, byval b8 as ubyte, byval a8 as ubyte=255)
  this.r=r8 : this.g=g8 : this.b=b8 : this.a=a8
end constructor

function RGBAColor.u8 as ubyte
  return (this.r and &HE0) or ((this.g and &HE0) shr 3) or ((this.b and &HC0) shr 6)
end function

function RGBAColor.u16 as ushort
  return ((this.r and &HF8) shl 8) or ((this.g and &HFC) shl 3) or ((this.b and &HF8) shr 3)
end function

type Matrix2D
  as single a, b, _
            c, d, _
            x, y
end type

type ClipRegion2D
  as integer xMin,yMin,xMax,yMax
end type

type Point2D
  as integer x,y
end type

type VideoMode
  as integer iWidth,iHeight,iBits,iBytes,iPitch
end type

type RenderContext
  declare destructor
  declare constructor(byval iWidth as integer=640, byval iHeight as integer=480, byval iBits as integer=32,byval bFullscreen as boolean=false)

  ' begin a new render frame (default clear the frame also)
  declare sub BeginFrame(byval bClearScreen as boolean=true)
  ' end current render frame and flip the result on screen
  declare sub EndFrame

  ' optional: set/get a clipping region default is the complete window/screen
  declare sub      setWindow(byval iLeft as integer,byval iTop as integer, byval iRight as integer, byval iBottom as integer, byval bDrawBoarder as boolean=false)
  declare function getLeft    as integer
  declare function getRight   as integer
  declare function getTop     as integer
  declare function getBottom  as integer
  ' right-left
  declare function getWidth   as integer
  ' bottom-top
  declare function getHeight  as integer
  ' left + width/2
  declare function getCenterX as integer
  ' top + height/2
  declare function getCenterY as integer

  ' set/get the drawing color (used by DrawPixel,DrawLine, DrawRectangle, DrawCircle, DrawElipse, DrawTriangle, DrawArc, DrawPie ...)
  declare sub      setColor(byval c as ulong)
  declare sub      setColor(byval r as ubyte, byval g as ubyte, byval b as ubyte, byval a as ubyte=&HFF)
  declare sub      setRed  (byval r as ubyte)
  declare sub      setGreen(byval g as ubyte)
  declare sub      setBlue (byval b as ubyte)
  declare sub      setAlpha(byval a as ubyte)
  declare function getColor as ulong
  declare function getRed   as ubyte
  declare function getGreen as ubyte
  declare function getBlue  as ubyte
  declare function getAlpha as ubyte

  ' set/get the fill color (used by FillRectangle, FillCircle, FillElipse, FillTriangle, FillPie, ...)
  declare sub      setFillColor(byval c as ulong)
  declare sub      setFillColor(byval r as ubyte, byval g as ubyte, byval b as ubyte, byval a as ubyte=&HFF)
  declare sub      setFillRed  (byval r as ubyte)
  declare sub      setFillGreen(byval g as ubyte)
  declare sub      setFillBlue (byval b as ubyte)
  declare sub      setFillAlpha(byval a as ubyte)
  declare function getFillColor as ulong
  declare function getFillRed   as ubyte
  declare function getFillGreen as ubyte
  declare function getFillBlue  as ubyte
  declare function getFillAlpha as ubyte

  ' set/get the boarder color (used by drawing any optinal boarder)
  declare sub      setBoarderColor(byval c as ulong)
  declare sub      setBoarderColor(byval r as ubyte, byval g as ubyte, byval b as ubyte, byval a as ubyte=&HFF)
  declare sub      setBoarderRed  (byval r as ubyte)
  declare sub      setBoarderGreen(byval g as ubyte)
  declare sub      setBoarderBlue (byval b as ubyte)
  declare sub      setBoarderAlpha(byval a as ubyte)
  declare function getBoarderColor as ulong
  declare function getBoarderRed   as ubyte
  declare function getBoarderGreen as ubyte
  declare function getBoarderBlue  as ubyte
  declare function getBoarderAlpha as ubyte

  ' set/get the clear color (used by BeginFrame(clearScreen=true))
  declare sub      setClearColor(byval c as ulong)
  declare sub      setClearColor(byval r as ubyte, byval g as ubyte, byval b as ubyte, byval a as ubyte=&HFF)
  declare sub      setClearRed  (byval r as ubyte)
  declare sub      setClearGreen(byval g as ubyte)
  declare sub      setClearBlue (byval b as ubyte)
  declare sub      setClearAlpha(byval a as ubyte)
  declare function getClearColor as ulong
  declare function getClearRed   as ubyte
  declare function getClearGreen as ubyte
  declare function getClearBlue  as ubyte
  declare function getClearAlpha as ubyte


  ' ########################
  ' # fast integer Drawing #
  ' ########################

  ' get pixel color in c return false if iX or iY are not in screen/window or current clipregion
  declare function GetPixel(byval iX as integer, byval iY as integer, byref c as ulong) as boolean
  ' set a pixel with current drawing color
  declare sub DrawPixel    (byval iX as integer, byval iY as integer)
  ' fill a region with FillColor the boundary is defined by BoundaryColor
  declare sub FillBoundary (byval iX as integer, byval iY as integer, byref FillColor as RGBAColor, byref BoundaryColor as RGBAColor)
  ' draw a horizontal line with current drawing color
  declare sub DrawHLine    (byval iX as integer, byval iY as integer, byval iX2 as integer)
  ' draw a vertical line with current drawing color
  declare sub DrawVLine    (byval iX as integer, byval iY as integer, byval iY2 as integer)
  ' draw a line with current drawing color
  declare sub DrawLine     (byval iX as integer, byval iY as integer, byval iX2 as integer, byval iY2 as integer)
  ' draw a rectangle with current drawing color
  declare sub DrawRectangle(byval iX as integer, byval iY as integer, byval iWidth as integer, byval iHeight as integer)
  ' fill a rectangle with current fill color optional draw the boarder with current boarder color
  declare sub FillRectangle(byval iX as integer, byval iY as integer, byval iWidth as integer, byval iHeight as integer,byval bBoarder as boolean=false)
  ' draw a circle with current drawing color
  declare sub DrawCircle   (byval iX as integer, byval iY as integer, byval iRadius as integer)
  ' fill a circle with current fill color optional draw the boarder with current boarder color
  declare sub FillCircle   (byval iX as integer, byval iY as integer, byval iRadius as integer, byval bBoarder as boolean=false)
  ' draw a elipse with current drawing color
  declare sub DrawElipse   (byval iX as integer, byval iY as integer, byval iXRadius as integer, byval iYRadius as integer)
  ' fill a elipse with current fill color optional draw the boarder with current boarder color
  declare sub FillElipse   (byval iX as integer, byval iY as integer, byval iXRadius as integer, byval iXRadius as integer, byval bBoarder as boolean=false)
  ' draw a triangle with current drawing color
  declare sub DrawTriangle (byval iX1 as integer, byval iY1 as integer, _
                            byval iX2 as integer, byval iY2 as integer, _
                            byval iX3 as integer, byval iY3 as integer)
  ' fill a triangle with current fill color optional draw the boarder with current boarder color
  declare sub FillTriangle (byval iX1 as integer, byval iY1 as integer, _
                            byval iX2 as integer, byval iY2 as integer, _
                            byval iX3 as integer, byval iY3 as integer, _
                            byval bBoarder as boolean=false)
  ' draw a bezier curve
  declare sub DrawCurve    (byval x1 as single, byval y1 as single, _
                            byval x2 as single, byval y2 as single, _
                            byval x3 as single, byval y3 as single, _
                            byval x4 as single, byval y4 as single)
  
  ' draw a part of an circle in current drawing color (for a closed or filled circle use DrawCircle or FillCircle)
  declare sub DrawArc(byval x as integer, byval y as integer, byval radius as integer, byval angleStart as integer, byval angleEnd as integer)
  ' draw a pie in current drawing color
  declare sub DrawPie(byval x as integer, byval y as integer, byval radius as integer, byval angleStart as integer, byval angleEnd as integer)
  ' fill a pie with current fill color optional draw the boarder with current boarder color
  declare sub FillPie(byval x as integer, byval y as integer, byval radius as integer, byval angleStart as integer, byval angleEnd as integer,byval bBoarder as boolean=false)

  ' #################
  ' # image drawing #
  ' #################
  declare function LoadImage(byval filename as const zstring ptr, byval DitherInPaletteMode as boolean=false) as any ptr

  declare function getImageWidth  (byval image as any ptr) as integer
  declare function getImageHeight (byval image as any ptr) as integer
  declare function getImageCenterX(byval image as any ptr) as single
  declare function getImageCenterY(byval image as any ptr) as single

  declare sub      DrawImage(byval image   as any ptr, _
                             byval xMidPos as integer, _
                             byval yMidPos as integer, _
                             byval xScale  as single=1, _
                             byval yScale  as single=1, _
                             byval angle   as integer=0)

  declare sub      DrawImage(byval image    as any ptr, _
                             byval xMidPos  as integer, _
                             byval yMidPos  as integer, _
                             byref ColorKey as RGBAColor, _
                             byval xScale   as single=1, _
                             byval yScale   as single=1, _
                             byval angle    as integer=0)

  ' ##################
  ' # vector drawing #
  ' ##################
  declare sub BeginPoints         ' begin a path of points
  declare sub EndPoints           ' draw the path of points with current drawing color

  declare sub BeginLines          ' begin a path of lines
  declare sub EndLines            ' draw the path of lines with current drawing color

  declare sub BeginLineLoop       ' begin a closed path of lines (last line will be connected to first line)
  declare sub EndLineLoop         ' draw the path with current drawing color

  declare sub BeginPolygon        ' begin a polygone path (concave)
  declare sub EndPolygon          ' fill the polygone path with current fill color
  ' under construction
  declare sub BeginComplexPolygon ' begin a complex polygone concave/convex wholes are posible also
  declare sub EndComplexPolygon   ' draw the complex polygone
  declare sub Gap                 ' separate two paths of a complex polygone 

  ' add a vertex (x,y position) to path the vertex will be transformed by current matrix
  declare sub AddVertex(byval x as single, byval y as single)

  ' optinal: add a pre transformed vertex (x,y position) to current path 
  declare sub AddTransformedVertex(byval iX as integer, byval iY as integer)

  ' add vertices on a bezier curve to the current path
  declare sub AddCurve(byval X0 as single, byval Y0 as single, _
                       byval X1 as single, byval Y1 as single, _
                       byval X2 as single, byval Y2 as single, _
                       byval X3 as single, byval Y3 as single)
  ' add vertices from an arc to the current path
  declare sub AddArc(byval x as single, byval y as single, byval r as single, byval angleStart as single, byval angleEnd as single)
  ' add vertices from an pie to the current path
  declare sub AddPie(byval x as single, byval y as single, byval r as single, byval angleStart as single, byval angleEnd as single)

  ' transformation
  declare sub      PushMatrix ' push the current matrix on stack
  declare sub      PopMatrix  ' pop last pushed matrix from stack and make it as current
  ' remove any rotation, scaling and translation from current matrix
  declare sub      Identiy
  ' add rotation to the current matrix
  declare sub      Rotate(byval degree as single)
  ' add scaling to the current matrix 
  declare sub      Scale(byval s as single)
  ' add x,y scaling to the current matrix 
  declare sub      Scale(byval sx as single,byval sy as single)
  ' add a x,y translation to the current matrix
  declare sub      Translate(byval tx as single,byval ty as single)

  ' tools
  ' return angle in degree between x,y and the origin (left,top)
  declare function AngleDegree(byval x as integer,byval y as integer) as integer
  ' return angle in degree between x1,y1 and x2,y2
  declare function AngleDegree(byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer) as integer

  ' ################
  ' # internal use #
  ' ################
  private:
  ' clip code
  const as integer cLEFT   = 1
  const as integer cRIGHT  = 2
  const as integer cBOTTOM = 4
  const as integer cTOP    = 8
  ' current vertex mode 
  const as integer vPOINTS   =  1
  const as integer vLINES    =  2
  const as integer vLINELOOP =  4
  const as integer vPOLYGON  =  8
  const as integer vCOMPLEX  = 16
  ' under construction
  declare sub      PushClipregion
  declare sub      PopClipregion
  ' later usefull to get mouse coords or screen coords in world coords
  declare function IverseMatrix(byref dst as Matrix2D, byref src as Matrix2D) as boolean
  ' not tested ATM.
  declare sub      Window2ViewportMatrix(byref dst as Matrix2D, _
                                         byval winX  as integer, byval winY  as integer, byval winW  as integer, byval winH  as integer, _
                                         byval viewX as integer, byval viewY as integer, byval viewW as integer, byval viewH as integer)
  ' multiply current matrix with a,b,c,d,e,x,y
  declare sub      MultMatrix(byval a as single, byval b as single, _
                              byval b as single, byval c as single, _
                              byval x as single, byval y as single)

  declare function TransformX (byval x as single, byval y as single) as single
  declare function TransformY (byval x as single, byval y as single) as single
  declare function TransformDX(byval x as single, byval y as single) as single
  declare function TransformDY(byval x as single, byval y as single) as single

  declare sub      Palette332
  declare function DitherPixel(byval iX as integer, byval iY as integer, byref c as RGBAColor) as ubyte
  declare function getClipcode(byval iX as integer, byval iY as integer) as integer
  declare sub      FillBoundary8 (byval iX as integer, byval iY as integer)
  declare sub      FillBoundary16(byval iX as integer, byval iY as integer)
  declare sub      FillBoundary32(byval iX as integer, byval iY as integer)
  declare sub      DrawLine8 (byval iX as integer, byval iY as integer, byval iX2 as integer, byval iY2 as integer)
  declare sub      DrawLine16(byval iX as integer, byval iY as integer, byval iX2 as integer, byval iY2 as integer)
  declare sub      DrawLine32(byval iX as integer, byval iY as integer, byval iX2 as integer, byval iY2 as integer)
  declare sub      FillPolygon8
  declare sub      FillPolygon16
  declare sub      FillPolygon32
  declare sub      FillPolygon
  declare sub      DrawPoints
  declare sub      DrawLines
  declare sub      DrawLineLoop

  declare function LoadDitherImage8(byval filename as const zstring ptr) as any ptr
  declare function LoadImage8 (byval filename as const zstring ptr) as any ptr
  declare function LoadImage16(byval filename as const zstring ptr) as any ptr
  declare function LoadImage32(byval filename as const zstring ptr) as any ptr

  declare sub      DrawImage8 (byval Pixels as ubyte ptr, byval iWidth as integer, byval iHeight as integer, byval iPitch as integer, _
                               byval xMidPos as integer, byval yMidPos as integer,byval xScale as single, byval yScale as single, byval angle as integer)
  declare sub      DrawImage16(byval Pixels as ushort ptr, byval iWidth as integer, byval iHeight as integer, byval iPitch as integer, _
                               byval xMidPos as integer, byval yMidPos as integer,byval xScale as single, byval yScale as single, byval angle as integer)
  declare sub      DrawImage32(byval Pixels as ulong ptr, byval iWidth as integer, byval iHeight as integer, byval iPitch as integer, _
                               byval xMidPos as integer, byval yMidPos as integer,byval xScale as single, byval yScale as single, byval angle as integer)

  declare sub      DrawImageColorkey8 (byval Pixels as ubyte ptr, byval iWidth as integer, byval iHeight as integer, byval iPitch as integer, _
                                    byval xMidPos as integer, byval yMidPos as integer,byval xScale as single, byval yScale as single, _
                                    byval angle as integer, byval ColorKey as ubyte)
  declare sub      DrawImageColorkey16(byval Pixels as ushort ptr, byval iWidth as integer, byval iHeight as integer, byval iPitch as integer, _
                                    byval xMidPos as integer, byval yMidPos as integer,byval xScale as single, byval yScale as single, _
                                    byval angle as integer, byval ColorKey as ushort)
  declare sub      DrawImageColorkey32(byval Pixels as ulong ptr, byval iWidth as integer, byval iHeight as integer, byval iPitch as integer, _
                                    byval xMidPos as integer, byval yMidPos as integer,byval xScale as single, byval yScale as single, _
                                    byval angle as integer, byval ColorKey as ulong)


  ' vector drawing
  as integer     nMaxPoints
  as Point2D ptr points
  as integer     nPoints
  as integer     vMode
  as integer     Gaps
  ' the colors
  as RGBAColor DrawingColor
  as RGBAColor FillColor
  as RGBAColor ClearColor
  as RGBAColor BoarderColor
  ' vide modes
  as VideoMode Desktop
  as VideoMode Current
  as integer iCenterX
  as integer iCenterY
  as ubyte  ptr pixels8
  as ushort ptr pixels16
  as ulong  ptr pixels32
  as boolean IsFullscreen
  ' flag for BeginFrame / EndFrame
  as boolean InFrame
 
  ' transforming
  const MATRIXSTACK_SIZE = 64
  as Matrix2D CurrentMatrix
  as Matrix2D MatrixStack(MATRIXSTACK_SIZE-1)
  as integer  MatrixStackIndex
  ' clipping
  const CLIPREGIONSTACK_SIZE = 32
  as ClipRegion2D CurrentClipregion
  as ClipRegion2D ClipregionStack(CLIPREGIONSTACK_SIZE-1)
  as integer  ClipregionStackIndex

  as boolean DrawBoarder
  ' recursive filler
  as integer MAX_BOUNDARIES
  as integer BoundaryIndex
  as ubyte   Boundaryc8
  as ushort  Boundaryc16
  as ulong   Boundaryc32
  as integer BoundaryCounter
  as integer ptr BoundaryX
  as integer ptr BoundaryY
  as integer nStack
  as ubyte  BoundaryColor8
  as ubyte  BoundaryFillColor8
  as ushort BoundaryColor16
  as ushort BoundaryFillColor16
  as ulong  BoundaryColor32
  as ulong  BoundaryFillColor32
end type

destructor RenderContext
  ' deallocate some stuff
  if points    then deallocate points
  if BoundaryX then deallocate BoundaryX
  if BoundaryY then deallocate BoundaryY
end destructor

constructor RenderContext(byval iWidth as integer, byval iHeight as integer, byval iBits as integer,byval bFullscreen as boolean)
  if screenptr() then screen 0
  with this.Desktop
    ScreenInfo(.iWidth,.iHeight,.iBits,.iBytes,.iPitch)
  end with
  if iBits < 15 then
    iBits =  8
  elseif iBits < 16 then
    iBits = 15
  elseif iBits < 24 then
    iBits = 16
  elseif iBits < 32 then
    iBits = 24
  else
    iBits = 32
  end if
  if bFullScreen=true then
    if iWidth<640 or iHeight<480 then
      iWidth=640 : iHeight=480
    end if
    dim as integer ModeBits(4)=>{8,15,16,24,32}
    dim as VideoMode VideoModes()
    dim as integer nModes,nFound=-1,nFoundSize=-1
    dim as boolean bitsFound
    for i as integer=0 to 4
      var mode = ScreenList(ModeBits(i))
      while mode<>0 andalso nFound=-1
        if ModeBits(i)=iBits then bitsFound=true
        redim preserve VideoModes(nModes)
        with VideoModes(nModes)
          .iWidth  = HiWord(mode)
          .iHeight = LoWord(mode)
          .iBits   = ModeBits(i)
          if .iWidth=iWidth andalso .iHeight=iHeight then
            nFoundSize = nModes
            if .iBits=iBits then nFound = nModes
          end if
        end with
        nModes+=1
        mode = ScreenList()
      wend
      if nFound<>-1 then exit for
    next

    if (nFound<>-1) then
      ' bits and size are found
      with VideoModes(nFound)
        isFullscreen=iif(ScreenRes(.iWidth,.iHeight,.iBits,2,1)=0,true,false)
      end with
    elseif (nFoundSize<>-1) then
      ' size are found
      with VideoModes(nFoundSize)
        isFullscreen=iif(ScreenRes(.iWidth,.iHeight,.iBits,2,1)=0,true,false)
      end with
    end if
  end if

  if isFullscreen=false then
    if iWidth<1 then
      iWidth=1
    elseif iWidth>this.Desktop.iWidth-32 then
      iWidth=this.Desktop.iWidth-32
    end if
    if iHeight<1 then
      iHeight=1
    elseif iHeight>this.Desktop.iHeight-32 then
      iHeight=this.Desktop.iHeight-32
    end if
    ScreenRes(iWidth,iHeight,iBits,2)
  end if

  screenset 1,0

  ' get current mode
  with current
    ScreenInfo(.iWidth,.iHeight,.iBits,.iBytes,.iPitch)
    this.iCenterX=.iWidth  shr 1
    this.iCenterY=.iHeight shr 1
  end with
  select case as const current.iBytes
  case 1 : this.pixels8 =screenptr() : this.Palette332()
  case 2 : this.pixels16=screenptr() : current.iPitch shr=1
  case 4 : this.pixels32=screenptr() : current.iPitch shr=2
  end select

  ' transform
  this.Identiy()
  ' clipping
  this.setWindow(0,0,this.current.iWidth-1,this.current.iHeight-1)

  this.setColor(255,255,255)
  this.setFillColor(32,32,32)
  this.setClearColor(128,128,255)
  this.setBoarderColor(128,128,128)

  ' recursive filler
  this.MAX_BOUNDARIES=(this.current.iWidth*this.current.iHeight)\2
  this.BoundaryX=allocate(sizeof(integer)*this.MAX_BOUNDARIES)
  this.BoundaryY=allocate(sizeof(integer)*this.MAX_BOUNDARIES)
  ' vector drawing
  this.nMaxPoints=16
  this.points=allocate(sizeof(Point2D)*this.nMaxPoints)
end constructor

function RenderContext.IverseMatrix(byref dst as Matrix2D, byref src as Matrix2D) as boolean
  dim as single determinant = src.a*src.d - src.b*src.c
  if (determinant > -1e-12) andalso (determinant < 1e-12) then return false
  determinant = 1/determinant
  dst.a =  src.d * determinant
  dst.b = -src.b * determinant
  dst.c = -src.c * determinant
  dst.d =  src.a * determinant
  dst.x = -src.x * dst.a - src.y * dst.c
  dst.y = -src.x * dst.b - src.y * dst.d
  return true
end function

sub RenderContext.Window2ViewportMatrix(byref dst as Matrix2D, _
                                        byval winX  as integer, byval winY  as integer, byval winW  as integer, byval winH as integer, _
                                        byval viewX as integer, byval viewY as integer, byval viewW as integer, byval viewH as integer)
  dim as single scaleX = viewW / winW
  dim as single scaleY = viewH / winH
  dst.a = scaleX : dst.b = 0.0
  dst.c = 0.0    : dst.d = scaleY
  dst.x = viewX - scaleX * winX
  dst.y = viewY - scaleY * winY
end sub


sub RenderContext.MultMatrix(byval a as single, byval b as single, _
                             byval c as single, byval d as single, _
                             byval x as single, byval y as single)
  dim as Matrix2D o
  o.a = a*this.CurrentMatrix.a + b*this.CurrentMatrix.c
  o.b = a*this.CurrentMatrix.b + b*this.CurrentMatrix.d
  
  o.c = c*this.CurrentMatrix.a + d*this.CurrentMatrix.c
  o.d = c*this.CurrentMatrix.b + d*this.CurrentMatrix.d
  
  o.x = x*this.CurrentMatrix.a + y*this.CurrentMatrix.c + this.CurrentMatrix.x
  o.y = x*this.CurrentMatrix.b + y*this.CurrentMatrix.d + this.CurrentMatrix.y
  this.CurrentMatrix = o
end sub


sub RenderContext.PushClipregion
  if (this.ClipregionStackIndex=this.CLIPREGIONSTACK_SIZE) then
    WindowTitle "error: PushClipregion stack overflow." : beep
  else
    this.ClipregionStack(this.ClipregionStackIndex) = this.CurrentClipregion
    this.ClipregionStackIndex+=1
  end if
end sub

sub RenderContext.PopClipregion
  if (this.ClipregionStackIndex=0) then
    WindowTitle "error: PopClipregion stack underflow." : beep
  else 
    this.ClipregionStackIndex-=1
    this.CurrentClipregion = this.ClipregionStack(this.ClipregionStackIndex)
  end if
end sub



function RenderContext.getClipCode(byval iX as integer, byval iY as integer) as integer
  dim as integer code
  if (iY > this.CurrentClipregion.yMax) then code or=this.cTOP
  if (iY < this.CurrentClipregion.yMin) then code or=this.cBOTTOM
  if (iX < this.CurrentClipregion.xMin) then code or=this.cLEFT
  if (iX > this.CurrentClipregion.xMax) then code or=this.cRIGHT
  return code
end function

sub RenderContext.Palette332() 
  dim as integer i,r,g,b 
  for i = 0 to 255 
    r=(((i shr 5) and &H07) * 255) / 7 
    g=(((i shr 2) and &H07) * 255) / 7 
    b=(((i shr 0) and &H03) * 255) / 3 
    palette i,r,g,b 
  next 
end sub 

function RenderContext.DitherPixel(byval x as integer, _
                                   byval y as integer, _
                                   byref c as RGBAColor) as ubyte 
  static as integer rtab(1,15) = { _
  {-16,  4, -1, 11,-14,  6, -3,  9,-15,  5, -2, 10,-13,  7, -4,  8}, _
  { 15, -5,  0,-12, 13, -7,  2,-10, 14, -6,  1,-11, 12, -8,  3, -9}} 
  static as integer gtab(1,15) = { _
  { 11,-15,  7, -3,  8,-14,  4, -2, 10,-16,  6, -4,  9,-13,  5, -1}, _
  {-12, 14, -8,  2, -9, 13, -5,  1,-11, 15, -7,  3,-10, 12, -6,  0}} 
  static as integer btab(1,15) = { _
  { -3,  9,-13,  7, -1, 11,-15,  5, -4,  8,-14,  6, -2, 10,-16,  4}, _
  {  2,-10, 12, -8,  0,-12, 14, -6,  3, -9, 13, -7,  1,-11, 15, -5}} 
  dim as integer r = c.r
  dim as integer g = c.g
  dim as integer b = c.b
  x and=&H1
  y and=&HF
  r+=rtab(x,y)
  g+=gtab(x,y)
  b+=btab(x,y)
  if (r > 254) then 
    r=224 ' &HE0 
  elseif r>0 then 
    r=r and &HE0 
  else 
    r=0 
  end if 
  if (g > 254) then 
    g=28 ' &HE0 shr 3 
  elseif g>0 then 
    g=(g and &HE0) shr 3 
  else 
    g=0 
  end if 
  if (b > 254) then 
    b=3 ' &HC0 shr 6 
  elseif b>0 then 
    b=(b and &HC0) shr 6 
  else 
    b=0 
  end if 
  r=r or g or b 
  return r
end function 


sub RenderContext.setWindow(byval iLeft   as integer, byval iTop    as integer, _
                            byval iRight  as integer, byval iBottom as integer, _
                            byval Boarder as boolean)

  if iLeft<0 then
    iLeft=0
  elseif iLeft>this.current.iWidth-2 then
    iLeft=this.current.iWidth-2
  end if
  if iRight>this.current.iWidth-1 then
    iRight=this.current.iWidth-1
  elseif iRight<=iLeft then
    iRight=iLeft+1
  end if

  if iTop<0 then
    iTop=0
  elseif iTop>this.current.iHeight-2 then
    iTop=this.current.iHeight-2
  end if
  if iBottom>this.current.iHeight-1 then
    iBottom=this.current.iHeight-1
  elseif iBottom<iTop+1 then
    iBottom=iTop+1
  end if

  this.DrawBoarder = Boarder
  this.CurrentClipregion.xMin = iLeft
  this.CurrentClipregion.yMin = iTop
  this.CurrentClipregion.xMax = iRight
  this.CurrentClipregion.yMax = iBottom
end sub

sub RenderContext.BeginFrame(byval bClear as boolean)
  if this.InFrame then this.EndFrame
  if bClear=true then
    swap this.FillColor.u32,this.ClearColor.u32
    this.FillRectangle(this.getLeft(),this.getTop(),this.getWidth(),this.getHeight())
    swap this.FillColor.u32,this.ClearColor.u32
  end if
  this.InFrame=true
end sub

sub RenderContext.EndFrame
  if this.InFrame=false then return
  if this.DrawBoarder then 
    swap this.DrawingColor.u32,this.BoarderColor.u32
    this.DrawRectangle(this.getLeft(),this.getTop(),this.getWidth(),this.getHeight())
    swap this.DrawingColor.u32,this.BoarderColor.u32
  end if
  flip
  this.InFrame=false
end sub

function RenderContext.getLeft as integer
  return this.CurrentClipregion.xMin
end function
function RenderContext.getRight as integer
  return this.CurrentClipregion.xMax
end function
function RenderContext.getTop as integer
  return this.CurrentClipregion.yMin
end function
function RenderContext.getBottom as integer
  return this.CurrentClipregion.yMax
end function

function RenderContext.getWidth   as integer
  return this.CurrentClipregion.xMax-this.CurrentClipregion.xMin
end function
function RenderContext.getHeight  as integer
  return this.CurrentClipregion.yMax-this.CurrentClipregion.yMin
end function
function RenderContext.getCenterX as integer
  return this.CurrentClipregion.xMin+this.getWidth shr 1
end function
function RenderContext.getCenterY as integer
  return this.CurrentClipregion.yMin+this.getHeight shr 1
end function

sub RenderContext.setColor(byval c as ulong)
  this.DrawingColor.u32=c
end sub
sub RenderContext.setColor(byval r as ubyte, byval g as ubyte, byval b as ubyte, byval a as ubyte)
  this.DrawingColor.u32=RGBA(r,g,b,a)
end sub
sub RenderContext.setRed(byval r as ubyte)
  this.DrawingColor.r=r
end sub
sub RenderContext.setGreen(byval g as ubyte)
  this.DrawingColor.g=g
end sub
sub RenderContext.setBlue(byval b as ubyte)
  this.DrawingColor.b=b
end sub
sub RenderContext.setAlpha(byval a as ubyte)
  this.DrawingColor.a=a
end sub
function RenderContext.getColor as ulong
  return this.DrawingColor.u32
end function 
function RenderContext.getRed as ubyte
  return this.DrawingColor.r
end function 
function RenderContext.getGreen as ubyte
  return this.DrawingColor.g
end function
function RenderContext.getBlue as ubyte
  return this.DrawingColor.b
end function
function RenderContext.getAlpha as ubyte
  return this.DrawingColor.a
end function

sub RenderContext.setFillColor(byval c as ulong)
  this.FillColor.u32=c
end sub
sub RenderContext.setFillColor(byval r as ubyte, byval g as ubyte, byval b as ubyte, byval a as ubyte)
  this.FillColor.u32=RGBA(r,g,b,a)
end sub
sub RenderContext.setFillRed  (byval r as ubyte)
  this.FillColor.r=r
end sub
sub RenderContext.setFillGreen(byval g as ubyte)
  this.FillColor.g=g
end sub
sub RenderContext.setFillBlue (byval b as ubyte)
  this.FillColor.b=b
end sub
sub RenderContext.setFillAlpha(byval a as ubyte)
  this.FillColor.a=a
end sub
function RenderContext.getFillColor as ulong
  return this.FillColor.u32
end function
function RenderContext.getFillRed as ubyte
  return this.FillColor.r
end function
function RenderContext.getFillGreen as ubyte
  return this.FillColor.g
end function
function RenderContext.getFillBlue as ubyte
  return this.FillColor.b
end function
function RenderContext.getFillAlpha as ubyte
  return this.FillColor.a
end function

sub RenderContext.setClearColor(byval c as ulong)
  this.ClearColor.u32=c
end sub
sub RenderContext.setClearColor(byval r as ubyte, byval g as ubyte, byval b as ubyte, byval a as ubyte)
  this.ClearColor.u32=RGBA(r,g,b,a)
end sub
sub RenderContext.setClearRed  (byval r as ubyte)
  this.ClearColor.r=r
end sub
sub RenderContext.setClearGreen(byval g as ubyte)
  this.ClearColor.g=g
end sub
sub RenderContext.setClearBlue (byval b as ubyte)
  this.ClearColor.b=b
end sub
sub RenderContext.setClearAlpha(byval a as ubyte)
  this.ClearColor.a=a
end sub
function RenderContext.getClearColor as ulong
  return this.ClearColor.u32
end function
function RenderContext.getClearRed as ubyte
  return this.ClearColor.r
end function
function RenderContext.getClearGreen as ubyte
  return this.ClearColor.g
end function
function RenderContext.getClearBlue as ubyte
  return this.ClearColor.b
end function
function RenderContext.getClearAlpha as ubyte
  return this.ClearColor.a
end function


sub RenderContext.setBoarderColor(byval c as ulong)
  this.BoarderColor.u32=c
end sub
sub RenderContext.setBoarderColor(byval r as ubyte, byval g as ubyte, byval b as ubyte, byval a as ubyte)
  this.BoarderColor.u32=RGBA(r,g,b,a)
end sub
sub RenderContext.setBoarderRed  (byval r as ubyte)
  this.BoarderColor.r=r
end sub
sub RenderContext.setBoarderGreen(byval g as ubyte)
  this.BoarderColor.g=g
end sub
sub RenderContext.setBoarderBlue (byval b as ubyte)
  this.BoarderColor.b=b
end sub
sub RenderContext.setBoarderAlpha(byval a as ubyte)
  this.BoarderColor.a=a
end sub
function RenderContext.getBoarderColor as ulong
  return this.BoarderColor.u32
end function
function RenderContext.getBoarderRed as ubyte
  return this.BoarderColor.r
end function
function RenderContext.getBoarderGreen as ubyte
  return this.BoarderColor.g
end function
function RenderContext.getBoarderBlue as ubyte
  return this.BoarderColor.b
end function
function RenderContext.getBoarderAlpha as ubyte
  return this.BoarderColor.a
end function


function RenderContext.GetPixel(byval iX as integer, _
                                byval iY as integer, _
                                byref c as ulong) as boolean
  if iX<this.CurrentClipregion.xMin then return false
  if iX>this.CurrentClipregion.xMax then return false
  if iY<this.CurrentClipregion.yMin then return false
  if iY>this.CurrentClipregion.yMax then return false
  select case as const this.current.iBytes
  case 1 : c = this.pixels8 [iY*this.current.iPitch+iX]
  case 2 : c = this.pixels16[iY*this.current.iPitch+iX]
  case 4 : c = this.pixels32[iY*this.current.iPitch+iX]
  end select
  return true
end function

#if 0
sub RenderContext.DrawPixel8(byval iX as integer, _
                             byval iY as integer) 
  this.pixels8[iY*this.current.iPitch+iX]=DrawingColor.u8
end sub
sub RenderContext.DrawPixel16(byval iX as integer, _
                              byval iY as integer) 
  this.pixels16[iY*this.current.iPitch+iX]=DrawingColor.u16
end sub
sub RenderContext.DrawPixel32(byval iX as integer, _
                             byval iY as integer) 
  this.pixels32[iY*this.current.iPitch+iX]=DrawingColor.u32
end sub
#endif

sub RenderContext.DrawPixel(byval iX as integer, _
                            byval iY as integer) 
  if iX<this.CurrentClipregion.xMin then return
  if iX>this.CurrentClipregion.xMax then return
  if iY<this.CurrentClipregion.yMin then return
  if iY>this.CurrentClipregion.yMax then return
  select case as const this.current.iBytes
  case 1 : this.pixels8 [iY*this.current.iPitch+iX]=DrawingColor.u8
  case 2 : this.pixels16[iY*this.current.iPitch+iX]=DrawingColor.u16
  case 4 : this.pixels32[iY*this.current.iPitch+iX]=DrawingColor.u32
  end select
end sub

sub RenderContext.FillBoundary8(byval iX as integer, byval iY as integer)
  if iX<this.CurrentClipregion.xMin then return
  if iX>this.CurrentClipregion.xMax then return
  if iY<this.CurrentClipregion.yMin then return
  if iY>this.CurrentClipregion.yMax then return
  this.BoundaryCounter+=sizeof(integer)*8
  if this.BoundaryCounter>1024*512 then
    if this.nStack<this.MAX_BOUNDARIES then
      this.BoundaryX[nStack]=iX
      this.BoundaryY[nStack]=iY
      this.nStack+=1
      'if this.nStack mod 1000=0 then windowTitle "push stack: " & this.nStack:sleep 100
    else
      windowTitle "out of stack space !"
    end if
    this.BoundaryCounter-=sizeof(integer)*8
    return
  end if
  this.BoundaryIndex=iY*this.current.iPitch+iX
  this.BoundaryC8=this.pixels8[this.BoundaryIndex]
  if (this.BoundaryC8<>this.BoundaryFillColor8) andalso (this.BoundaryC8<>this.BoundaryColor8) then
    'this.DrawPixel(iX,iY)
    this.pixels8[this.BoundaryIndex]=this.BoundaryFillColor8
    this.FillBoundary8(iX+1,iY)
    this.FillBoundary8(iX,iY+1)
    this.FillBoundary8(iX,iY-1)
    this.FillBoundary8(iX-1,iY)
  end if
  this.BoundaryCounter-=sizeof(integer)*8
end sub
sub RenderContext.FillBoundary16(byval iX as integer, byval iY as integer)
  if iX<this.CurrentClipregion.xMin then return
  if iX>this.CurrentClipregion.xMax then return
  if iY<this.CurrentClipregion.yMin then return
  if iY>this.CurrentClipregion.yMax then return
  this.BoundaryCounter+=sizeof(integer)*8
  if this.BoundaryCounter>1024*512 then
    if this.nStack<this.MAX_BOUNDARIES then
      this.BoundaryX[nStack]=iX
      this.BoundaryY[nStack]=iY
      this.nStack+=1
      'if this.nStack mod 1000=0 then windowTitle "push stack: " & this.nStack:sleep 100
    else
      windowTitle "out of stack space !"
    end if
    this.BoundaryCounter-=sizeof(integer)*8
    return
  end if
  this.BoundaryIndex = iY*this.current.iPitch+iX
  this.BoundaryC16 = this.pixels16[this.BoundaryIndex]
  if (this.BoundaryC16<>this.BoundaryFillColor16) andalso (this.BoundaryC16<>this.BoundaryColor16) then
    'this.DrawPixel(iX,iY)
    this.pixels16[this.BoundaryIndex]=this.BoundaryFillColor16
    this.FillBoundary16(iX+1,iY)
    this.FillBoundary16(iX,iY+1)
    this.FillBoundary16(iX,iY-1)
    this.FillBoundary16(iX-1,iY)
  end if
  this.BoundaryCounter-=sizeof(integer)*8
end sub
sub RenderContext.FillBoundary32(byval iX as integer, byval iY as integer)
  if iX<this.CurrentClipregion.xMin then return
  if iX>this.CurrentClipregion.xMax then return
  if iY<this.CurrentClipregion.yMin then return
  if iY>this.CurrentClipregion.yMax then return
  this.BoundaryCounter+=sizeof(integer)*8
  if this.BoundaryCounter>1024*512 then
    if this.nStack<this.MAX_BOUNDARIES then
      this.BoundaryX[nStack]=iX
      this.BoundaryY[nStack]=iY
      this.nStack+=1
      'if this.nStack mod 1000=0 then windowTitle "push stack: " & this.nStack:sleep 100
    else
      'windowTitle "out of stack space !"
    end if
    this.BoundaryCounter-=sizeof(integer)*8
    return
  end if
  this.BoundaryIndex=iY*this.current.iPitch+iX
  this.BoundaryC32=this.pixels32[this.BoundaryIndex]
  if (this.BoundaryC32<>this.BoundaryFillColor32) andalso (this.BoundaryC32<>this.BoundaryColor32) then
    this.pixels32[this.BoundaryIndex]=this.BoundaryFillColor32
    this.FillBoundary32(iX+1,iY)
    this.FillBoundary32(iX,iY+1)
    this.FillBoundary32(iX,iY-1)
    this.FillBoundary32(iX-1,iY)
  end if
  this.BoundaryCounter-=sizeof(integer)*8
end sub

sub RenderContext.FillBoundary(byval iX as integer, byval iY as integer,byref f as RGBAColor,byref b as RGBAColor)
  this.BoundaryCounter=0
  this.nStack=0
  select case as const this.current.iBytes
  case 1
    this.BoundaryColor8    =b.u8
    this.BoundaryFillColor8=f.u8
    this.FillBoundary8(iX,iY)
    while this.nStack>0
      this.nStack-=1 : this.FillBoundary8(this.BoundaryX[this.nStack],this.BoundaryY[this.nStack])
    wend
  case 2
    this.BoundaryColor16    =b.u16
    this.BoundaryFillColor16=f.u16
    this.FillBoundary16(iX,iY)
    while this.nStack>0
      this.nStack-=1 : this.FillBoundary16(this.BoundaryX[this.nStack],this.BoundaryY[this.nStack])
    wend
  case 4
    this.BoundaryColor32    =b.u32
    this.BoundaryFillColor32=f.u32
    this.FillBoundary32(iX,iY)
    while this.nStack>0
      this.nStack-=1 : this.FillBoundary32(this.BoundaryX[this.nStack],this.BoundaryY[this.nStack])
    wend
  end select
end sub

sub RenderContext.DrawHLine(byval iX as integer, byval iY as integer, byval iX2 as integer)
  if iY<this.CurrentClipregion.yMin then return
  if iY>this.CurrentClipregion.yMax then return
  if iX>iX2 then swap iX,iX2
  if iX>this.CurrentClipregion.xMax then
    return
  elseif ix2<this.CurrentClipregion.xMin then
    return
  end if
  if iX <this.CurrentClipregion.xMin then iX =this.CurrentClipregion.xMin
  if iX2>this.CurrentClipregion.xMax then iX2=this.CurrentClipregion.xMax
  dim as integer w=iX2-iX
  select case as const this.current.iBytes
  case 1
    dim as ubyte c = this.DrawingColor.u8
    dim as ubyte ptr s = @this.pixels8[iY*this.current.iPitch+iX]
    if w<1 then *s=c : return
    dim as ubyte ptr e = s+w+1
    while s<e : *s=c : s+=1 : wend
  case 2
    dim as ushort c = this.DrawingColor.u16
    dim as ushort ptr s = @this.pixels16[iY*this.current.iPitch+iX]
    if w<1 then *s=c : return
    dim as ushort ptr e = s+w+1
    while s<e : *s=c : s+=1 : wend
  case 4
    dim as ulong c = this.DrawingColor.u32
    dim as ulong ptr s = @this.pixels32[iY*this.current.iPitch+iX]
    if w<1 then *s=c : return
    dim as ulong ptr e = s+w+1
    while s<e : *s=c : s+=1 : wend
  end select
end sub

sub RenderContext.DrawVLine(byval iX as integer, byval iY as integer, byval iY2 as integer)
  if iX<this.CurrentClipregion.xMin then
    return
  elseif iX>this.CurrentClipregion.xMax then
    return
  end if
  if iY>iY2 then swap iY,iY2
  if iY>this.CurrentClipregion.yMax then
    return
  elseif iY2<this.CurrentClipregion.yMin then
    return
  end if
  if iY <this.CurrentClipregion.yMin then iY =this.CurrentClipregion.yMin
  if iY2>this.CurrentClipregion.yMax then iY2=this.CurrentClipregion.yMax
  dim as integer h=iY2-iY
  if h<1 then this.DrawPixel(iX,iY) : return

  select case as const this.current.iBytes
  case 1
    dim as ubyte c = this.DrawingColor.u8
    dim as ubyte ptr s = @this.pixels8[iY*this.current.iPitch+iX]
    dim as ubyte ptr e = s+h*this.current.iPitch
    while s<e : *s=c : s+=this.current.iPitch : wend
  case 2
    dim as ushort c = this.DrawingColor.u16
    dim as ushort ptr s = @this.pixels16[iY*this.current.iPitch+iX]
    dim as ushort ptr e = s+h*this.current.iPitch
    while s<e : *s=c : s+=this.current.iPitch : wend
  case 4
    dim as ulong c = this.DrawingColor.u32
    dim as ulong ptr s = @this.pixels32[iY*this.current.iPitch+iX]
    dim as ulong ptr e = s+h*this.current.iPitch
    while s<e : *s=c : s+=this.current.iPitch : wend
  end select
end sub

sub RenderContext.DrawLine8(byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer)
  dim as ubyte ptr s=any, e=any
  dim as integer pixelStep=any, rowStep=any
  dim as integer xError=any, yError=any, cError=any, dx=any, dy=any
  dim as ubyte c=this.DrawingColor.u8 
  dx = x2-x1
  dy = y2-y1
  if dx=0 andalso dy=0 then
    ' plot a pixel
    this.pixels8[y1*this.current.iPitch+x1]=c
  elseif dx=0 then
    ' draw a vertical line from [s]tart to [e]nd (top to bottom)
    if y1>y2 then swap y1,y2
    s = this.pixels8 + y1*this.current.iPitch + x1 : e = s + this.current.iPitch*(y2-y1)
    while s<e : *s=c : s+=this.current.iPitch : wend
  elseif dy=0 then
    ' draw a horizontal line from [s]tart to [e]nd (left to right)
    if x1>x2 then swap x1,x2
    s = this.pixels8 + y1*this.current.iPitch + x1 : e = s + (x2-x1)
    while s<e : *s=c : s+=1 : wend
  else
    ' draw a diagonal line from [s]tart to [e]nd
    if dx<0 then xError = -dx else xError = dx
    if dy<0 then yError = -dy else yError = dy 
    pixelStep = iif( dx<0,-1    ,1     ) ' one pixel left or right
    rowStep   = iif( dy<0,-this.current.iPitch,this.current.iPitch) ' one row   up   or down
    s = this.pixels8 + y1*this.current.iPitch + x1    ' start of line
    e = s + yError*rowStep + xError*pixelStep ' end   of line
    xError shl=1 : yError shl=1
    if (xError>yError) then ' x - loop
      cError = yError-(xError shr 1) 
    else                    ' y - loop
      cError = xError-(yError shr 1) 
      swap pixelStep,rowStep  ' swap pixel and row step
      swap xError,yError      ' swap x,y error
    end if  
    rowStep+=pixelStep
    while s<>e
      *s=c : if (cError < 0) then s += pixelStep else s += rowStep : cError -= xError 
      cError += yError
    wend
  end if
end sub

sub RenderContext.DrawLine16(byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer)
  dim as ushort ptr s=any, e=any
  dim as integer pixelStep=any, rowStep=any
  dim as integer xError=any, yError=any, cError=any, dx=any, dy=any
  dim as ushort c=this.DrawingColor.u16 
  dx = x2-x1
  dy = y2-y1
  if dx=0 andalso dy=0 then
    ' plot a pixel
    this.pixels16[y1*this.current.iPitch+x1]=c
  elseif dx=0 then
    ' draw a vertical line from [s]tart to [e]nd (top to bottom)
    if y1>y2 then swap y1,y2
    s = this.pixels16 + y1*this.current.iPitch + x1 : e = s + this.current.iPitch*(y2-y1)
    while s<e : *s=c : s+=this.current.iPitch : wend
  elseif dy=0 then
    ' draw a horizontal line from [s]tart to [e]nd (left to right)
    if x1>x2 then swap x1,x2
    s = this.pixels16 + y1*this.current.iPitch + x1 : e = s + (x2-x1)
    while s<e : *s=c : s+=1 : wend
  else
    ' draw a diagonal line from [s]tart to [e]nd
    if dx<0 then xError = -dx else xError = dx
    if dy<0 then yError = -dy else yError = dy 
    pixelStep = iif( dx<0,-1    ,1     ) ' one pixel left or right
    rowStep   = iif( dy<0,-this.current.iPitch,this.current.iPitch) ' one row   up   or down
    s = this.pixels16 + y1*this.current.iPitch + x1    ' start of line
    e = s + yError*rowStep + xError*pixelStep ' end   of line
    xError shl=1 : yError shl=1
    if (xError>yError) then ' x - loop
      cError = yError-(xError shr 1) 
    else                    ' y - loop
      cError = xError-(yError shr 1) 
      swap pixelStep,rowStep  ' swap pixel and row step
      swap xError,yError      ' swap x,y error
    end if  
    rowStep+=pixelStep
    while s<>e
      *s=c : if (cError < 0) then s += pixelStep else s += rowStep : cError -= xError 
      cError += yError
    wend
  end if
end sub

sub RenderContext.DrawLine32(byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer)
  dim as ulong ptr s=any, e=any
  dim as integer pixelStep=any, rowStep=any
  dim as integer xError=any, yError=any, cError=any, dx=any, dy=any
  dim as ulong c=this.DrawingColor.u32 
  dx = x2-x1
  dy = y2-y1
  if dx=0 andalso dy=0 then
    ' plot a pixel
    this.pixels32[y1*this.current.iPitch+x1]=c
  elseif dx=0 then
    ' draw a vertical line from [s]tart to [e]nd (top to bottom)
    if y1>y2 then swap y1,y2
    s = this.pixels32 + y1*this.current.iPitch + x1 : e = s + this.current.iPitch*(y2-y1)
    while s<e : *s=c : s+=this.current.iPitch : wend
  elseif dy=0 then
    ' draw a horizontal line from [s]tart to [e]nd (left to right)
    if x1>x2 then swap x1,x2
    s = this.pixels32 + y1*this.current.iPitch + x1 : e = s + (x2-x1)
    while s<e : *s=c : s+=1 : wend
  else
    ' draw a diagonal line from [s]tart to [e]nd
    if dx<0 then xError = -dx else xError = dx
    if dy<0 then yError = -dy else yError = dy 
    pixelStep = iif( dx<0,-1    ,1     ) ' one pixel left or right
    rowStep   = iif( dy<0,-this.current.iPitch,this.current.iPitch) ' one row   up   or down
    s = this.pixels32 + y1*this.current.iPitch + x1    ' start of line
    e = s + yError*rowStep + xError*pixelStep ' end   of line
    xError shl=1 : yError shl=1
    if (xError>yError) then ' x - loop
      cError = yError-(xError shr 1) 
    else                    ' y - loop
      cError = xError-(yError shr 1) 
      swap pixelStep,rowStep  ' swap pixel and row step
      swap xError,yError      ' swap x,y error
    end if  
    rowStep+=pixelStep
    while s<>e
      *s=c : if (cError < 0) then s += pixelStep else s += rowStep : cError -= xError 
      cError += yError
    wend
  end if
end sub

sub RenderContext.DrawLine(byval iX as integer, byval iY as integer, byval iX2 as integer, byval iY2 as integer)
  dim as integer x=any,y=any,temp=any
  dim as integer code1=any,code2=any
  dim as integer dx=(iX2-iX),dy=(iY2-iY)
  if (dx=0) andalso (dy=0) then
    this.DrawPixel(iX,iY) : return
  elseif (dx=0) then
    this.DrawVLine(iX,iY,iY2) : return
  elseif (dy=0) then
    this.DrawHLine(iX,iY,iX2) : return
  else
    ' get clipcodes
    code1=this.getClipCode(iX,iY)
    code2=this.getClipCode(iX2,iY2)
    if (code1=0) andalso (code2=0) then 
      ' complete inside
      select case as const this.current.iBytes
      case 1 : this.DrawLine8 (iX,iY,iX2,iY2)
      case 2 : this.DrawLine16(iX,iY,iX2,iY2)
      case 4 : this.DrawLine32(iX,iY,iX2,iY2)
      end select
      return
    elseif (code1 and code2)<>0 then
      ' complete outside
      return
    end if
  end if
    
  do 
    if (code1=0) andalso (code2=0) then 
      ' complete inside
      select case as const this.current.iBytes
      case 1 : this.DrawLine8 (iX,iY,iX2,iY2)
      case 2 : this.DrawLine16(iX,iY,iX2,iY2)
      case 4 : this.DrawLine32(iX,iY,iX2,iY2)
      end select
      return
    elseif (code1 and code2)<>0 then
      ' complete outside
      return
    else
      dim as single m = (iY2-iY) / (iX2-iX)
      if (code1=0) then
        temp = code2
      else 
        temp = code1
      end if
      if (temp and cTOP) then
        x = iX + (CurrentClipregion.yMax-iY)/m : y = CurrentClipregion.yMax
      elseif(temp and cBOTTOM) then
        x = iX + (CurrentClipregion.yMin-iY)/m : y = CurrentClipregion.yMin
      elseif (temp and cLEFT) then
        x = CurrentClipregion.xMin : y = iY+ m*(CurrentClipregion.xMin-iX)
      elseif (temp and cRIGHT) then
        x = CurrentClipregion.xMax : y = iY+ m*(CurrentClipregion.xMax-iX)
      end if
      if (temp = code1) then
        iX = x  : iY = y  : code1 = getClipCode(iX,iY)
      else
        iX2 = x : iY2 = y : code2 = getClipCode(iX2,iY2)
      end if
    end if
  loop
end sub

sub RenderContext.DrawRectangle(byval iX as integer, byval iY as integer, byval iWidth as integer, byval iHeight as integer)
  if (iWidth<1) andalso (iHeight<1) then 
    this.DrawPixel(iX,iY) : return
  elseif (iWidth<1) then
    this.DrawVLine(iX,iY,iY+iHeight) : return
  elseif (iHeight<1) then
    this.DrawHLine(iX,iY,iX+iWidth)  : return
  end if
  dim as integer iX2=iX+iWidth
  dim as integer iY2=iY+iHeight
  this.DrawHLine(iX ,iY ,iX2)
  this.DrawHLine(iX ,iY2,iX2)
  this.DrawVLine(iX ,iY ,iY2)
  this.DrawVLine(iX2,iY ,iY2)
end sub
sub RenderContext.FillRectangle(byval iX as integer, byval iY as integer, byval iWidth as integer, byval iHeight as integer,byval Boarder as boolean)
  dim as integer iX2=iX+iWidth
  dim as integer iY2=iY+iHeight
  swap this.DrawingColor.u32,this.FillColor.u32
  for y as integer = iY to iY2
    this.DrawHLine(ix,y,iX2)
  next
  swap this.DrawingColor.u32,this.FillColor.u32
  if Boarder then
    swap this.DrawingColor.u32,this.BoarderColor.u32
    this.DrawRectangle(iX,iY,iWidth,iHeight)
    swap this.DrawingColor.u32,this.BoarderColor.u32
  end if
end sub

sub RenderContext.DrawCircle(byval cx as integer, byval cy as integer, byval r as integer)
  this.DrawElipse(cx,cy,r,r)
end sub
sub RenderContext.FillCircle(byval cx as integer, byval cy as integer, byval r as integer,byval bBoarder as boolean)
  this.FillElipse(cx,cy,r,r,bBoarder)
end sub

sub RenderContext.DrawElipse(byval cx as integer, byval cy as integer, byval rx as integer, byval ry as integer)
  if (rx<1) andalso (ry<1) then
    this.DrawPixel(cx,cy) : return
  elseif (rx<1) then
    this.DrawVLine(cx,cy-ry,cy+ry) : return
  elseif (ry<1) then
    this.DrawHLine(cx-rx,cy,cx+rx) : return
  end if

  if (cx+rx) < 0 then return
  if (cx-rx) > this.current.iWidth-1 then return
  if (cy+ry) < 0 then return
  if (cy-ry) > this.current.iHeight-1 then return

  ' Region 1
  dim as integer rx2=rx*rx
  dim as integer ry2=ry*ry
  dim as integer rx22=rx2 shl 1
  dim as integer ry22=ry2 shl 1
  dim as single  p=ry2 - rx2*ry + rx2*0.25
  dim as integer x,y=ry
  while (ry22*x <= rx22*y)
    if (p < 0) then
      p += ry22*x + ry2
    else
      y -= 1
      p += ry22*x - rx22*y-ry2
    end if
     this.DrawPixel(cx-x,cy-y)
     this.DrawPixel(cx+x,cy-y)
     this.DrawPixel(cx-x,cy+y)
     this.DrawPixel(cx+x,cy+y)
     x += 1
  wend
  ' Region 2
  p=ry2*(x+0.5)*(x+0.5) + rx2*(y-1)*(y-1) - rx2*ry2
  while (y > 0)
    y-=1
    if (p <= 0) then
      x += 1
      p += ry22*x - rx22*y + rx2
    else
      p -= rx22*y + rx2
    end if
    this.DrawPixel(cx-x,cy-y)
    this.DrawPixel(cx+x,cy-y)
    this.DrawPixel(cx-x,cy+y)
    this.DrawPixel(cx+x,cy+y)
  wend
end sub

sub RenderContext.FillElipse(byval cx as integer, byval cy as integer, byval rx as integer, byval ry as integer,byval bBoarder as boolean)
  if (rx<1) andalso (ry<1) then
    swap this.DrawingColor.u32,this.FillColor.u32
    this.DrawPixel(cx,cy) 
    swap this.DrawingColor.u32,this.FillColor.u32
    return
  elseif (rx<1) then
    swap this.DrawingColor.u32,this.FillColor.u32
    this.DrawVLine(cx,cy-ry,cy+ry)
    swap this.DrawingColor.u32,this.FillColor.u32
    return
  elseif (ry<1) then
    swap this.DrawingColor.u32,this.FillColor.u32
    this.DrawHLine(cx-rx,cy,cx+rx)
    swap this.DrawingColor.u32,this.FillColor.u32
    return
  end if
  if (cx+rx) < 0 then return
  if (cx-rx) > this.current.iWidth-1 then return
  if (cy+ry) < 0 then return
  if (cy-ry) > this.current.iHeight-1 then return

  swap this.DrawingColor.u32,this.FillColor.u32

  ' Region 1
  dim as integer rx2=rx*rx
  dim as integer ry2=ry*ry
  dim as integer rx22=rx2 shl 1
  dim as integer ry22=ry2 shl 1
  dim as single  p=ry2 - rx2*ry + rx2*0.25
  dim as integer x,y=ry
  while (ry22*x <= rx22*y)
    x += 1
    if (p < 0) then
      p += ry22*x + ry2
    else
      y -= 1
      p += ry22*x - rx22*y-ry2
    end if
    this.DrawHLine(cx-x,cy-y,cx+x)
    this.DrawHLine(cx-x,cy+y,cx+x)
  wend
  ' Region 2
  p=ry2*(x+0.5)*(x+0.5) + rx2*(y-1)*(y-1) - rx2*ry2
  while (y > 0)
    y-=1
    if (p <= 0) then
      x += 1
      p += ry22*x - rx22*y + rx2
    else
      p -= rx22*y + rx2
    end if
    this.DrawHLine(cx-x,cy-y,cx+x)
    this.DrawHLine(cx-x,cy+y,cx+x)
  wend
  swap this.DrawingColor.u32,this.FillColor.u32
  if bBoarder then
    swap this.DrawingColor.u32,this.BoarderColor.u32
    DrawElipse(cx,cy,rx,ry)
    swap this.DrawingColor.u32,this.BoarderColor.u32
  end if
end sub

' #################
' # image drawing #
' #################
' Load a RGBA 32-bit image and convert it to 8-bit
function RenderContext.LoadImage8(byval filename as const zstring ptr) as any ptr
  var img32 = LoadRGBAFile(filename)
  if img32=0 then return 0
  dim as RGBAColor c
  dim as integer w,h,spitch,dpitch
  dim as ulong ptr s
  imageinfo img32,w,h,,spitch,s : spitch shr=2
  var img8 = ImageCreate(w,h,0,8)
  dim as ubyte ptr d
  imageinfo img8,,,,dpitch,d 
  for y as integer =1 to h
    dim as integer i
    for x as integer =0 to w-1
      c.u32=s[x] : d[x] = c.u8
    next
    s+=spitch ' next src row
    d+=dpitch ' next dst row
  next
  ImageDestroy img32
  return img8
end function

' Load a RGBA 32-bit image and convet and dither it to to 8-bit
function RenderContext.LoadDitherImage8(byval filename as const zstring ptr) as any ptr
  var img32 = LoadRGBAFile(filename)
  if img32=0 then return 0
  dim as RGBAColor c
  dim as integer w,h,spitch,dpitch
  dim as ulong ptr s
  imageinfo img32,w,h,,spitch,s : spitch shr=2
  var img8 = ImageCreate(w,h,0,8)
  dim as ubyte ptr d
  imageinfo img8 , , ,,dpitch,d 
  for y as integer =0 to h -1
    for x as integer =0 to w-1
      c.u32=s[x] : d[x] = this.DitherPixel(x,y,c)
    next
    s+=spitch ' next src row
    d+=dpitch ' next dst row
  next
  ImageDestroy img32
  return img8
end function

' Load a RGBA 32-bit image and convert it to 16-bit
function RenderContext.LoadImage16(byval filename as const zstring ptr) as any ptr

  var img32 = LoadRGBAFile(filename)
  if img32=0 then return 0
  dim as RGBAColor c
  dim as integer w,h,spitch,dpitch
  dim as ulong ptr s
  imageinfo img32,w,h,,spitch,s : spitch shr=2
  var img16 = ImageCreate(w,h,0,16)
  dim as ushort ptr d
  imageinfo img16, , ,,dpitch,d : dpitch shr=1
  for y as integer =1 to h
    for x as integer =0 to w-1
      c.u32=s[x] : d[x] = c.u16
    next
    s+=spitch ' next src row
    d+=dpitch ' next dst row
  next
  ImageDestroy img32

  return img16
end function

' load a 32-bit RGBA image
function RenderContext.LoadImage32(byval filename as const zstring ptr) as any ptr
  return LoadRGBAFile(filename)
end function

function RenderContext.LoadImage(byval filename as const zstring ptr,byval dither as boolean) as any ptr
  select case as const this.Current.iBytes
  case 1 : if dither then return LoadDitherImage8(filename) else return this.LoadImage8(filename)
  case 2 : return this.LoadImage16(filename)
  case 4 : return this.LoadImage32(filename)
  end select
end function

function RenderContext.getImageWidth(byval image as any ptr) as integer
  dim as integer w
  if image=0 then return 0
  if imageinfo(image,w) then return 0
  return w
end function
function RenderContext.getImageHeight(byval image as any ptr) as integer
  dim as integer h
  if image=0 then return 0
  if imageinfo(image,,h) then return 0
  return h
end function
function RenderContext.getImageCenterX(byval image as any ptr) as single
  return this.getImageWidth(image)*0.5
end function
function RenderContext.getImageCenterY(byval image as any ptr) as single
  return this.getImageHeight(image)*0.5
end function



' from Multiput by D.J.Peters (Joshy)
type FP16 ' fixed point 16:16
  union
  type
    as ushort l
    as  short h
  end type
  as integer v
  end union
end type

#define _ADD_ 0 ' increment a value
#define _CMP_ 1 ' compare values
#define _SET_ 2 ' set a value

#define _XScreen_  0
#define _YScreen_  1
#define _UTexture_ 2
#define _VTexture_ 3

#define _LeftIndex_    0
#define _RightIndex_   1

#define _CurrentIndex_ 0
#define _NextIndex_    1

#define _EdgeXStart_ 0
#define _EdgeUStart_ 1
#define _EdgeVStart_ 2
#define _EdgeXStep_  3
#define _EdgeUStep_  4
#define _EdgeVStep_  5

#macro defDrawImageColorkey(_name_,_source_,_type_)
Sub _name_ (byval ImagePixels as _type_ ptr, _
            byval ImageWidth  as integer, _
            byval ImageHeight as integer, _
            byval ImagePitch  as integer, _
            byval xMidPos     as integer, _
            byval yMidPos     as integer, _
            byval xScale      as single , _
            byval yScale      as single , _
            byval angle       as integer, _
            byval colorkey    as _type_)
  Dim As Integer i=any,yStart=any,yEnd=any,xStart=any,xEnd=any
  Dim As Integer CNS(1,1)=any 'Counters
  Dim As Integer ACS(1,2)=any '_ADD_ _COMPARE_ and _SET_
  Dim As Single fPoints(3,3)=any,fEdges(2,6)=any,fLength=any,fUSlope=any,fVSlope=any
  Dim As FP16 U=any,V=any,US=any,VS=any
  Dim as _type_ ptr RenderRow=any,s=any,e=any
  dim as _type_ c
  If xScale < 0.001 Then xScale=0.001
  If yScale < 0.001 Then yScale=0.001

  fPoints(0,_XScreen_)=-ImageWidth/2 * xScale
  fPoints(1,_XScreen_)= ImageWidth/2 * xScale
  fPoints(2,_XScreen_)= fPoints(1,_XScreen_)
  fPoints(3,_XScreen_)= fPoints(0,_XScreen_)

  fPoints(0,_YScreen_)=-ImageHeight/2 * yScale
  fPoints(1,_YScreen_)= fPoints(0,_YScreen_)
  fPoints(2,_YScreen_)= ImageHeight/2 * yScale
  fPoints(3,_YScreen_)= fPoints(2,_YScreen_)

  fPoints(0,_UTexture_)=0
  fPoints(1,_UTexture_)= ImageWidth
  fPoints(2,_UTexture_)= fPoints(1,_UTexture_)
  fPoints(3,_UTexture_)=0
 
  fPoints(0,_VTexture_)=0
  fPoints(1,_VTexture_)=0
  fPoints(2,_VTexture_)= ImageHeight
  fPoints(3,_VTexture_)= fPoints(2,_VTexture_)

  If angle<>0 Then
    var rad = angle * DEG2RAD
    var co = cos(rad)
    var si = sin(rad)
    For i=0 To 3
      var x = fPoints(i,_XScreen_)*co - fPoints(i,_YScreen_)*si
      var y = fPoints(i,_XScreen_)*si + fPoints(i,_YScreen_)*co
      fPoints(i,_XScreen_) = x
      fPoints(i,_YScreen_) = y
    Next
  End If
  yStart=2^28:yEnd=-yStart:xStart=yStart:xEnd=yEnd
 
  ' get min max
  For i=0 To 3
    fPoints(i,_XScreen_)=Int(fPoints(i,_XScreen_)+xMidPos)
    fPoints(i,_YScreen_)=Int(fPoints(i,_YScreen_)+yMidPos)
    If fPoints(i,_YScreen_)<yStart Then yStart=fPoints(i,_YScreen_):CNS(_LeftIndex_,_CurrentIndex_)=i
    If fPoints(i,_YScreen_)>yEnd   Then yEnd  =fPoints(i,_YScreen_)
    If fPoints(i,_XScreen_)<xStart Then xStart=fPoints(i,_XScreen_)
    If fPoints(i,_XScreen_)>xEnd   Then xEnd  =fPoints(i,_XScreen_)
  Next
 'windowtitle "" & yStart & "," & yEnd & "," & xStart & "," & xEnd
  If yEnd-yStart<2                          then exit sub
  If xEnd-xStart<2                          then exit sub
  If yEnd   < this.currentclipregion.yMin+1 then exit sub
  If xEnd   < this.currentclipregion.xMin+1 then exit sub
  If yStart > this.currentclipregion.yMax-1 then exit sub
  If xStart > this.currentclipregion.xMax-1 then exit sub

  ACS(_LeftIndex_ ,_ADD_)=-1:ACS(_LeftIndex_ ,_CMP_)=-1:ACS(_LeftIndex_ ,_SET_)=3
  ACS(_RightIndex_,_ADD_)= 1:ACS(_RightIndex_,_CMP_)= 4:ACS(_RightIndex_,_SET_)=0

  ' share the same highest point
  CNS(_RightIndex_,_CurrentIndex_)=CNS(_LeftIndex_,_CurrentIndex_)
 
  RenderRow = _source_ + ystart*this.Current.iPitch
  ' loop from Top to Bottom
  While yStart<yEnd
    'Scan Left and Right edges together
    For i=_LeftIndex_ To _RightIndex_
      ' bad to read but fast and short ;-)
      If yStart=fPoints(CNS(i,_CurrentIndex_),_YScreen_) Then
        CNS(i,_NextIndex_)=CNS(i,_CurrentIndex_)+ACS(i,_ADD_)
        If CNS(i,_NextIndex_)=ACS(i,_CMP_) Then CNS(i,_NextIndex_)=ACS(i,_SET_)
        While fPoints(CNS(i,_CurrentIndex_),_YScreen_) = fPoints(CNS(i,_NextIndex_),_YScreen_)
          CNS(i,_CurrentIndex_)=CNS(i,_NextIndex_)
          CNS(i,_NextIndex_   )=CNS(i,_CurrentIndex_)+ACS(i,_ADD_)
          If CNS(i,_NextIndex_)=ACS(i,_CMP_) Then CNS(i,_NextIndex_)=ACS(i,_SET_)
        Wend
        fEdges(i,_EdgeXStart_) = fPoints(CNS(i,_CurrentIndex_),_XScreen_)
        fEdges(i,_EdgeUStart_) = fPoints(CNS(i,_CurrentIndex_),_UTexture_)
        fEdges(i,_EdgeVStart_) = fPoints(CNS(i,_CurrentIndex_),_VTexture_)
        fLength  = fPoints(CNS(i,_NextIndex_),_YScreen_) - fPoints(CNS(i,_CurrentIndex_),_YScreen_)
        If fLength <> 0.0 Then
          fLength=1/fLength
          fEdges(i,_EdgeXStep_) = fPoints(CNS(i,_NextIndex_),_XScreen_ )-fEdges(i,_EdgeXStart_):fEdges(i,_EdgeXStep_)*=fLength
          fEdges(i,_EdgeUStep_) = fPoints(CNS(i,_NextIndex_),_UTexture_)-fEdges(i,_EdgeUStart_):fEdges(i,_EdgeUStep_)*=fLength
          fEdges(i,_EdgeVStep_) = fPoints(CNS(i,_NextIndex_),_VTexture_)-fEdges(i,_EdgeVStart_):fEdges(i,_EdgeVStep_)*=fLength
        End If
        CNS(i,_CurrentIndex_)=CNS(i,_NextIndex_)
      End If
    Next

    If (yStart<this.CurrentClipRegion.yMin) then goto NextScanLine
    xStart=fEdges(_LeftIndex_ ,_EdgeXStart_):If xStart>this.currentclipregion.xMax then goto NextScanLine
    xEnd  =fEdges(_RightIndex_,_EdgeXStart_):If xEnd  <this.CurrentClipRegion.xMin then goto NextScanLine
    If (xStart = xEnd  ) then goto NextScanLine
    if (xEnd   < xStart) then goto NextScanLine

    fLength=1/(xEnd-xStart)
    fUSlope=fEdges(_RightIndex_,_EdgeUStart_)-fEdges(_LeftIndex_,_EdgeUStart_):fUSlope*=fLength
    fVSlope=fEdges(_RightIndex_,_EdgeVStart_)-fEdges(_LeftIndex_,_EdgeVStart_):fVSlope*=fLength
    If xStart<this.CurrentClipRegion.xMin Then
      fLength=this.CurrentClipRegion.xMin-xStart
      U.v=(fEdges(_LeftIndex_,_EdgeUStart_)+fUSlope*fLength)*&HFFFF
      V.v=(fEdges(_LeftIndex_,_EdgeVStart_)+fVSlope*fLength)*&HFFFF
      xStart = this.CurrentClipRegion.xMin
    Else
      U.v=fEdges(_LeftIndex_,_EdgeUStart_)*&HFFFF
      V.v=fEdges(_LeftIndex_,_EdgeVStart_)*&HFFFF
    End If
    If u.v<0 Then u.v=0
    If v.v<0 Then v.v=0
    US.v=fUSlope*&HFFFF
    VS.v=fVSlope*&HFFFF

    If xEnd>this.currentclipregion.xMax Then xEnd=this.currentclipregion.xMax
    s=RenderRow + xStart
    e=s+(xEnd-xStart)
    While s<e
      c=*(ImagePixels+V.h*ImagePitch+U.h)
      if c<>colorkey then *s=c
      V.v+=VS.v : U.v+=US.v : s+=1
    Wend
  NextScanLine:
    yStart+=1 : If yStart>this.CurrentClipRegion.yMax Then exit while
    RenderRow += this.current.iPitch
    fEdges(_LeftIndex_ ,_EdgeXStart_)+=fEdges(_LeftIndex_ ,_EdgeXStep_)
    fEdges(_LeftIndex_ ,_EdgeUStart_)+=fEdges(_LeftIndex_ ,_EdgeUStep_)
    fEdges(_LeftIndex_ ,_EdgeVStart_)+=fEdges(_LeftIndex_ ,_EdgeVStep_)
    fEdges(_RightIndex_,_EdgeXStart_)+=fEdges(_RightIndex_,_EdgeXStep_)
    fEdges(_RightIndex_,_EdgeUStart_)+=fEdges(_RightIndex_,_EdgeUStep_)
    fEdges(_RightIndex_,_EdgeVStart_)+=fEdges(_RightIndex_,_EdgeVStep_)
  Wend
End Sub
#endmacro
defDrawImageColorkey(RenderContext.DrawImageColorkey8,this.pixels8,ubyte)
defDrawImageColorkey(RenderContext.DrawImageColorkey16,this.pixels16,ushort)
defDrawImageColorkey(RenderContext.DrawImageColorkey32,this.pixels32,ulong)

#undef defDrawImageColorkey

#macro defDrawImage(_name_,_source_,_type_)
Sub _name_ (byval ImagePixels as _type_ ptr, _
           byval ImageWidth  as integer, _
           byval ImageHeight as integer, _
           byval ImagePitch  as integer, _
           byval xMidPos     as integer, _
           byval yMidPos     as integer, _
           byval xScale      as single , _
           byval yScale      as single , _
           byval angle       as integer)
  Dim As Integer i=any,yStart=any,yEnd=any,xStart=any,xEnd=any
  Dim As Integer CNS(1,1)=any 'Counters
  Dim As Integer ACS(1,2)=any '_ADD_ compare and _SET_
  Dim As Single fPoints(3,3)=any,fEdges(2,6)=any,fLength=any,fUSlope=any,fVSlope=any
  Dim As FP16 U=any,V=any,US=any,VS=any
  Dim as _type_ ptr RenderRow=any,s=any,e=any

  If xScale < 0.001 Then xScale=0.001
  If yScale < 0.001 Then yScale=0.001

  fPoints(0,_XScreen_)=-ImageWidth/2 * xScale
  fPoints(1,_XScreen_)= ImageWidth/2 * xScale
  fPoints(2,_XScreen_)= fPoints(1,_XScreen_)
  fPoints(3,_XScreen_)= fPoints(0,_XScreen_)

  fPoints(0,_YScreen_)=-ImageHeight/2 * yScale
  fPoints(1,_YScreen_)= fPoints(0,_YScreen_)
  fPoints(2,_YScreen_)= ImageHeight/2 * yScale
  fPoints(3,_YScreen_)= fPoints(2,_YScreen_)

  fPoints(0,_UTexture_)=0
  fPoints(1,_UTexture_)= ImageWidth
  fPoints(2,_UTexture_)= fPoints(1,_UTexture_)
  fPoints(3,_UTexture_)=0
 
  fPoints(0,_VTexture_)=0
  fPoints(1,_VTexture_)=0
  fPoints(2,_VTexture_)= ImageHeight
  fPoints(3,_VTexture_)= fPoints(2,_VTexture_)

  If angle<>0 Then
    var rad = angle * DEG2RAD
    var co = cos(rad)
    var si = sin(rad)
    For i=0 To 3
      var x = fPoints(i,_XScreen_)*co - fPoints(i,_YScreen_)*si
      var y = fPoints(i,_XScreen_)*si + fPoints(i,_YScreen_)*co
      fPoints(i,_XScreen_) = x
      fPoints(i,_YScreen_) = y
    Next
  End If
  yStart=30^2:yEnd=-yStart:xStart=yStart:xEnd=yEnd
 
  ' get min max
  For i=0 To 3
    fPoints(i,_XScreen_)=Int(fPoints(i,_XScreen_)+xMidPos)
    fPoints(i,_YScreen_)=Int(fPoints(i,_YScreen_)+yMidPos)
    If fPoints(i,_YScreen_)<yStart Then yStart=fPoints(i,_YScreen_):CNS(_LeftIndex_,_CurrentIndex_)=i
    If fPoints(i,_YScreen_)>yEnd   Then yEnd  =fPoints(i,_YScreen_)
    If fPoints(i,_XScreen_)<xStart Then xStart=fPoints(i,_XScreen_)
    If fPoints(i,_XScreen_)>xEnd   Then xEnd  =fPoints(i,_XScreen_)
  Next
 'windowtitle "" & yStart & "," & yEnd & "," & xStart & "," & xEnd
  If yEnd-yStart<2                          then exit sub
  If xEnd-xStart<2                          then exit sub
  If yEnd   < this.currentclipregion.yMin+1 then exit sub
  If xEnd   < this.currentclipregion.xMin+1 then exit sub
  If yStart > this.currentclipregion.yMax-1 then exit sub
  If xStart > this.currentclipregion.xMax-1 then exit sub

  ACS(_LeftIndex_ ,_ADD_)=-1:ACS(_LeftIndex_ ,_CMP_)=-1:ACS(_LeftIndex_ ,_SET_)=3
  ACS(_RightIndex_,_ADD_)= 1:ACS(_RightIndex_,_CMP_)= 4:ACS(_RightIndex_,_SET_)=0

  ' share the same highest point
  CNS(_RightIndex_,_CurrentIndex_)=CNS(_LeftIndex_,_CurrentIndex_)
 
  RenderRow = _source_ + ystart*this.Current.iPitch
  ' loop from Top to Bottom
  While yStart<yEnd
    'Scan Left and Right edges together
    For i=_LeftIndex_ To _RightIndex_
      ' bad to read but fast and short ;-)
      If yStart=fPoints(CNS(i,_CurrentIndex_),_YScreen_) Then
        CNS(i,_NextIndex_)=CNS(i,_CurrentIndex_)+ACS(i,_ADD_)
        If CNS(i,_NextIndex_)=ACS(i,_CMP_) Then CNS(i,_NextIndex_)=ACS(i,_SET_)
        While fPoints(CNS(i,_CurrentIndex_),_YScreen_) = fPoints(CNS(i,_NextIndex_),_YScreen_)
          CNS(i,_CurrentIndex_)=CNS(i,_NextIndex_)
          CNS(i,_NextIndex_   )=CNS(i,_CurrentIndex_)+ACS(i,_ADD_)
          If CNS(i,_NextIndex_)=ACS(i,_CMP_) Then CNS(i,_NextIndex_)=ACS(i,_SET_)
        Wend
        fEdges(i,_EdgeXStart_) = fPoints(CNS(i,_CurrentIndex_),_XScreen_)
        fEdges(i,_EdgeUStart_) = fPoints(CNS(i,_CurrentIndex_),_UTexture_)
        fEdges(i,_EdgeVStart_) = fPoints(CNS(i,_CurrentIndex_),_VTexture_)
        fLength  = fPoints(CNS(i,_NextIndex_),_YScreen_) - fPoints(CNS(i,_CurrentIndex_),_YScreen_)
        If fLength <> 0.0 Then
          fLength=1/fLength
          fEdges(i,_EdgeXStep_) = fPoints(CNS(i,_NextIndex_),_XScreen_ )-fEdges(i,_EdgeXStart_):fEdges(i,_EdgeXStep_)*=fLength
          fEdges(i,_EdgeUStep_) = fPoints(CNS(i,_NextIndex_),_UTexture_)-fEdges(i,_EdgeUStart_):fEdges(i,_EdgeUStep_)*=fLength
          fEdges(i,_EdgeVStep_) = fPoints(CNS(i,_NextIndex_),_VTexture_)-fEdges(i,_EdgeVStart_):fEdges(i,_EdgeVStep_)*=fLength
        End If
        CNS(i,_CurrentIndex_)=CNS(i,_NextIndex_)
      End If
    Next

    If (yStart<this.CurrentClipRegion.yMin) then goto NextScanLine
    xStart=fEdges(_LeftIndex_ ,_EdgeXStart_):If xStart>this.currentclipregion.xMax then goto NextScanLine
    xEnd  =fEdges(_RightIndex_,_EdgeXStart_):If xEnd  <this.CurrentClipRegion.xMin then goto NextScanLine
    If (xStart = xEnd  ) then goto NextScanLine
    if (xEnd   < xStart) then goto NextScanLine

    fLength=1/(xEnd-xStart)
    fUSlope=fEdges(_RightIndex_,_EdgeUStart_)-fEdges(_LeftIndex_,_EdgeUStart_):fUSlope*=fLength
    fVSlope=fEdges(_RightIndex_,_EdgeVStart_)-fEdges(_LeftIndex_,_EdgeVStart_):fVSlope*=fLength
    If xStart<this.CurrentClipRegion.xMin Then
      fLength=this.CurrentClipRegion.xMin-xStart
      U.v=(fEdges(_LeftIndex_,_EdgeUStart_)+fUSlope*fLength)*&HFFFF
      V.v=(fEdges(_LeftIndex_,_EdgeVStart_)+fVSlope*fLength)*&HFFFF
      xStart = this.CurrentClipRegion.xMin
    Else
      U.v=fEdges(_LeftIndex_,_EdgeUStart_)*&HFFFF
      V.v=fEdges(_LeftIndex_,_EdgeVStart_)*&HFFFF
    End If
    If u.v<0 Then u.v=0
    If v.v<0 Then v.v=0
    US.v=fUSlope*&HFFFF
    VS.v=fVSlope*&HFFFF

    If xEnd>this.currentclipregion.xMax Then xEnd=this.currentclipregion.xMax
    s=RenderRow + xStart
    e=s+(xEnd-xStart)
    While s<e
      *s=*(ImagePixels+V.h*ImagePitch+U.h)
      V.v+=VS.v : U.v+=US.v : s+=1
    Wend
  NextScanLine:
    yStart+=1 : If yStart>this.CurrentClipRegion.yMax Then exit while
    RenderRow += this.current.iPitch
    fEdges(_LeftIndex_ ,_EdgeXStart_)+=fEdges(_LeftIndex_ ,_EdgeXStep_)
    fEdges(_LeftIndex_ ,_EdgeUStart_)+=fEdges(_LeftIndex_ ,_EdgeUStep_)
    fEdges(_LeftIndex_ ,_EdgeVStart_)+=fEdges(_LeftIndex_ ,_EdgeVStep_)
    fEdges(_RightIndex_,_EdgeXStart_)+=fEdges(_RightIndex_,_EdgeXStep_)
    fEdges(_RightIndex_,_EdgeUStart_)+=fEdges(_RightIndex_,_EdgeUStep_)
    fEdges(_RightIndex_,_EdgeVStart_)+=fEdges(_RightIndex_,_EdgeVStep_)
  Wend
End Sub
#endmacro

defDrawImage(RenderContext.DrawImage8,this.pixels8,ubyte)
defDrawImage(RenderContext.DrawImage16,this.pixels16,ushort)
defDrawImage(RenderContext.DrawImage32,this.pixels32,ulong)

#undef defDrawImage

#undef _ADD_
#undef _CMP_
#undef _SET_
#undef _XScreen_
#undef _YScreen_
#undef _UTexture_
#undef _VTexture_
#undef _LeftIndex_
#undef _RightIndex_
#undef _CurrentIndex_
#undef _NextIndex_
#undef _EdgeXStart_
#undef _EdgeUStart_
#undef _EdgeVStart_
#undef _EdgeXStep_
#undef _EdgeUStep_
#undef _EdgeVStep_

sub RenderContext.DrawImage(byval image      as any ptr, _
                            byval xMidPos    as integer, _
                            byval yMidPos    as integer, _
                            byval xScale     as single, _
                            byval yScale     as single, _
                            byval angle      as integer)
  if image=0 then return
  dim as integer iWidth,iHeight,iBytes,iPitch
  dim as any ptr Pixels
  if ImageInfo(image,iWidth,iHeight,iBytes,iPitch,Pixels) then return
  if iBytes<>this.Current.iBytes then return
  select case as const this.Current.iBytes
  case 1 :                this.DrawImage8 (pixels,iWidth,iHeight,iPitch,xMidPos,yMidPos,xScale,yScale,angle)
  case 2 : iPitch shr=1 : this.DrawImage16(pixels,iWidth,iHeight,iPitch,xMidPos,yMidPos,xScale,yScale,angle)
  case 4 : iPitch shr=2 : this.DrawImage32(pixels,iWidth,iHeight,iPitch,xMidPos,yMidPos,xScale,yScale,angle)
  'windowTitle "" & pixels & "," & iWidth & "," & iHeight & "," & iPitch & "," & xMidPos & "," & yMidPos & "," & xScale & "," & yScale & "," & angle
  end select
end sub

sub RenderContext.DrawImage(byval image      as any ptr, _
                            byval xMidPos    as integer, _
                            byval yMidPos    as integer, _
                            byref TransColor as RGBAColor, _
                            byval xScale     as single=1, _
                            byval yScale     as single=1, _
                            byval angle      as integer=0)
  if image=0 then return
  dim as integer iWidth,iHeight,iBytes,iPitch
  dim as any ptr Pixels
  if ImageInfo(image,iWidth,iHeight,iBytes,iPitch,Pixels) then return
  if iBytes<>this.Current.iBytes then return
  select case as const this.Current.iBytes
  case 1 :                this.DrawImageColorkey8 (pixels,iWidth,iHeight,iPitch,xMidPos,yMidPos,xScale,yScale,angle,TransColor.u8)
  case 2 : iPitch shr=1 : this.DrawImageColorkey16(pixels,iWidth,iHeight,iPitch,xMidPos,yMidPos,xScale,yScale,angle,TransColor.u16)
  case 4 : iPitch shr=2 : this.DrawImageColorkey32(pixels,iWidth,iHeight,iPitch,xMidPos,yMidPos,xScale,yScale,angle,TransColor.u32)
  end select
end sub

sub RenderContext.FillPolygon8
  dim as integer f =any,t =any,b=any,l =any,r=any
  dim as integer lv=any,lc=any,nlc=any,rc=any,nrc=any
  dim as integer lx=any,ls=any,rx=any,rs=any,cl=any,cr=any
  dim as ubyte     c8=any '     8 bit color
  dim as ubyte ptr s=any,e=any,row=any
  dim as Point2D ptr p=this.points
  ' last vertex 0 based
  lv=this.nPoints-1
  ' get min max coords
  t= 100000000 : b=-t : l= t :  r=-t
  for nc as integer=0 to lv
    if p[nc].y<t then t=p[nc].y:f=nc ' top ([f]ound=most top point index)
    if p[nc].y>b then b=p[nc].y      ' bottom
    if p[nc].x<l then l=p[nc].x      ' left
    if p[nc].x>r then r=p[nc].x      ' right
  next
  if (r-l)<1 then exit sub  ' 0 pixels width
  if (b-t)<1 then exit sub  ' 0 pixels height

  ' clip 
  if l>this.CurrentClipregion.xMax then exit sub  ' left is outside
  if r<this.CurrentClipregion.xMin then exit sub  ' right is outside
  if t>this.CurrentClipregion.yMax then exit sub  ' top is outside
  if b<this.CurrentClipregion.yMin then exit sub  ' bottom is outside
  if b>this.CurrentClipregion.yMax then b=this.CurrentClipregion.yMax ' clip bottom
  
  ' left and next left counter
  lc=f:nlc=lc-1:if nlc<0 then nlc=lv
  ' right and next right counter
  rc=f:nrc=rc+1:if nrc>lv then nrc=0
  ' in case of a CCS polygon
  'if p[nlc].x>=p[nrc].x then exit sub
  
  b+=1
  ' first row
  row=this.pixels8+t*this.current.iPitch
  c8=this.FillColor.u8
  while t<b ' from top to bottom
    if t=p[lc].y then
      while p[lc].y=p[nlc].y
        lc=nlc:nlc-=1:if nlc<0 then nlc=lv
      wend
      lx = p[lc].x shl 16
      ls = ((p[nlc].x-p[lc].x) shl 16)/(p[nlc].y-p[lc].y)
      lc = nlc
    end if
    if t=p[rc].y then
      while p[rc].y=p[nrc].y
        rc=nrc:nrc+=1:if nrc>lv then nrc=0
      wend
      rx = p[rc].x shl 16
      rs = ((p[nrc].x-p[rc].x) shl 16)/(p[nrc].y-p[rc].y)
      rc = nrc
    end if
    ' top inside ?
    if t>this.CurrentClipregion.yMin-1 then
      l=lx shr 16 ' most left  pixel
      r=rx shr 16 ' most right pixel
      'if l>r then swap l,r
      ' left and right inside ?
      if l<this.CurrentClipregion.xMax andalso r>this.CurrentClipregion.xMin then
        if l<this.CurrentClipregion.xMin then l=this.CurrentClipregion.xMin ' clip left 
        if r>this.CurrentClipregion.xMax then r=this.CurrentClipregion.xMax ' clip right
        s=row+l
        e=row+r+1
        while s<e : *s=c8 : s+=1 : wend
      end if
    end if
    row+=this.Current.iPitch ' next row
    lx+=ls ' left  x + left  step (in fixed point)
    rx+=rs ' right x + right step (in fixed point)
    t+=1   ' increment loop counter (top to bottom)
  wend
end sub

sub RenderContext.FillPolygon16
  dim as integer f =any,t =any,b=any,l =any,r=any
  dim as integer lv=any,lc=any,nlc=any,rc=any,nrc=any
  dim as integer lx=any,ls=any,rx=any,rs=any,cl=any,cr=any
  dim as ushort c16=any ' 16 bit color
  dim as ushort ptr s=any,e=any,row=any
  dim as Point2D ptr p=this.points
  ' last vertex 0 based
  lv=this.nPoints-1
  ' get min max coords
  t= 100000000 : b=-t : l= t : r=-t
  for nc as integer=0 to lv
    if p[nc].y<t then t=p[nc].y:f=nc ' top ([f]ound=most top point index)
    if p[nc].y>b then b=p[nc].y      ' bottom
    if p[nc].x<l then l=p[nc].x      ' left
    if p[nc].x>r then r=p[nc].x      ' right
  next
  if (r-l)<1 then exit sub  ' 0 pixels width
  if (b-t)<1 then exit sub  ' 0 pixels height

  ' clip 
  if l>this.CurrentClipregion.xMax then exit sub  ' left is outside
  if r<this.CurrentClipregion.xMin then exit sub  ' right is outside
  if t>this.CurrentClipregion.yMax then exit sub  ' top is outside
  if b<this.CurrentClipregion.yMin then exit sub  ' bottom is outside
  if b>this.CurrentClipregion.yMax then b=this.CurrentClipregion.yMax ' clip bottom
  
  ' left and next left counter
  lc=f:nlc=lc-1:if nlc<0 then nlc=lv
  ' right and next right counter
  rc=f:nrc=rc+1:if nrc>lv then nrc=0
  ' in case of a CCS polygon
  'if p[nlc].x>=p[nrc].x then exit sub
  
  b+=1
  ' first row
  row=this.pixels16+t*this.current.iPitch
  c16=this.FillColor.u16
  while t<b ' from top to bottom
    if t=p[lc].y then
      while p[lc].y=p[nlc].y
        lc=nlc:nlc-=1:if nlc<0 then nlc=lv
      wend
      lx = p[lc].x shl 16
      ls = ((p[nlc].x-p[lc].x) shl 16)/(p[nlc].y-p[lc].y)
      lc = nlc
    end if
    if t=p[rc].y then
      while p[rc].y=p[nrc].y
        rc=nrc:nrc+=1:if nrc>lv then nrc=0
      wend
      rx = p[rc].x shl 16
      rs = ((p[nrc].x-p[rc].x) shl 16)/(p[nrc].y-p[rc].y)
      rc = nrc
    end if
    ' top inside ?
    if t>this.CurrentClipregion.yMin-1 then
      l=lx shr 16 ' most left  pixel
      r=rx shr 16 ' most right pixel
      'if l>r then swap l,r
      ' left and right inside ?
      if l<this.CurrentClipregion.xMax andalso r>this.CurrentClipregion.xMin then
        if l<this.CurrentClipregion.xMin then l=this.CurrentClipregion.xMin ' clip left 
        if r>this.CurrentClipregion.xMax then r=this.CurrentClipregion.xMax ' clip right
        s=row+l
        e=row+r+1
        while s<e : *s=c16 : s+=1 : wend
      end if
    end if
    row+=this.Current.iPitch ' next row
    lx+=ls ' left  x + left  step (in fixed point)
    rx+=rs ' right x + right step (in fixed point)
    t+=1   ' increment loop counter (top to bottom)
  wend
end sub

sub RenderContext.FillPolygon32 
  dim as integer f =any,t =any,b=any,l =any,r=any
  dim as integer lv=any,lc=any,nlc=any,rc=any,nrc=any
  dim as integer lx=any,ls=any,rx=any,rs=any,cl=any,cr=any
  dim as ulong c32=any '     32 bit color
  dim as ulong ptr s=any,e=any,row=any
  dim as Point2D ptr p=this.points
  ' last vertex 0 based
  lv=this.nPoints-1
 ' get min max coords
  t= 100000000 : b=-t : l=t : r=-t 
  for nc as integer=0 to lv
    if p[nc].y<t then t=p[nc].y:f=nc ' top ([f]ound=most top point index)
    if p[nc].y>b then b=p[nc].y      ' bottom
    if p[nc].x<l then l=p[nc].x      ' left
    if p[nc].x>r then r=p[nc].x      ' right
  next
  if (r-l)<1 then exit sub  ' 0 pixels width
  if (b-t)<1 then exit sub  ' 0 pixels height

  ' clip 
  if l>this.CurrentClipregion.xMax then exit sub  ' left is outside
  if r<this.CurrentClipregion.xMin then exit sub  ' right is outside
  if t>this.CurrentClipregion.yMax then exit sub  ' top is outside
  if b<this.CurrentClipregion.yMin then exit sub  ' bottom is outside
  if b>this.CurrentClipregion.yMax then b=this.CurrentClipregion.yMax ' clip bottom
  
  ' left and next left counter
  lc=f:nlc=lc-1:if nlc<0 then nlc=lv
  ' right and next right counter
  rc=f:nrc=rc+1:if nrc>lv then nrc=0
  ' in case of a CCS polygon
  'if p[nlc].x>=p[nrc].x then exit sub
  
  b+=1
  ' first row
  row=this.pixels32+t*this.current.iPitch
  c32=this.FillColor.u32
  while t<b ' from top to bottom
    if t=p[lc].y then
      while p[lc].y=p[nlc].y
        lc=nlc:nlc-=1:if nlc<0 then nlc=lv
      wend
      lx = p[lc].x shl 16
      ls = ((p[nlc].x-p[lc].x) shl 16)/(p[nlc].y-p[lc].y)
      lc = nlc
    end if
    if t=p[rc].y then
      while p[rc].y=p[nrc].y
        rc=nrc:nrc+=1:if nrc>lv then nrc=0
      wend
      rx = p[rc].x shl 16
      rs = ((p[nrc].x-p[rc].x) shl 16)/(p[nrc].y-p[rc].y)
      rc = nrc
    end if
    ' top inside ?
    if t>CurrentClipregion.yMin-1 then
      l=lx shr 16 ' most left  pixel
      r=rx shr 16 ' most right pixel
      'if l>r then swap l,r
      ' left and right inside ?
      if l<this.CurrentClipregion.xMax andalso r>this.CurrentClipregion.xMin then
        if l<this.CurrentClipregion.xMin then l=this.CurrentClipregion.xMin ' clip left 
        if r>this.CurrentClipregion.xMax then r=this.CurrentClipregion.xMax ' clip right
        s=row+l
        e=row+r+1
        while s<e : *s=c32 : s+=1 : wend
      end if
    end if
    row+=this.Current.iPitch ' next row
    lx+=ls ' left  x + left  step (in fixed point)
    rx+=rs ' right x + right step (in fixed point)
    t+=1   ' increment loop counter (top to bottom)
  wend
end sub


sub RenderContext.DrawTriangle(byval x0 as integer, byval y0 as integer, _
                               byval x1 as integer, byval y1 as integer, _
                               byval x2 as integer, byval y2 as integer)
  var oldnPoints = this.nPoints
  var oldPoints  = this.Points
  this.points    = @this.Points[oldnPoints]
  this.nPoints   = 4
  this.Points[0].x=x0:this.Points[0].y=y0
  this.Points[1].x=x1:this.Points[1].y=y1
  this.Points[2].x=x2:this.Points[2].y=y2
  this.points[3].x=x0:this.points[3].y=y0
  this.DrawLines()
  this.Points  = oldPoints
  this.nPoints = oldnPoints
end sub

sub RenderContext.FillTriangle(byval x0 as integer, byval y0 as integer, _
                               byval x1 as integer, byval y1 as integer, _
                               byval x2 as integer, byval y2 as integer, _
                               byval bDrawBoarder as boolean)
  var oldnPoints = this.nPoints
  var oldPoints  = this.Points
  this.points    = @this.Points[oldnPoints]
  this.nPoints   = 3
  this.Points[0].x=x0:this.Points[0].y=y0
  this.Points[1].x=x1:this.Points[1].y=y1
  this.Points[2].x=x2:this.Points[2].y=y2
  this.FillPolygon()
  if bDrawBoarder=true then
    this.nPoints = 4
    this.points[3].x=x0:this.points[3].y=y0
    swap this.DrawingColor.u32,this.BoarderColor.u32
    this.DrawLines()
    swap this.DrawingColor.u32,this.BoarderColor.u32
  end if
  this.Points  = oldPoints
  this.nPoints = oldnPoints
end sub


sub RenderContext.DrawCurve(byval x0 as single, byval y0 as single, _
                            byval x1 as single, byval y1 as single, _
                            byval x2 as single, byval y2 as single, _
                            byval x3 as single, byval y3 as single)
  dim as single a = abs((x0-x2)*(y3-y1)-(y0-y2)*(x3-x1))
  dim as single b = abs((x0-x3)*(y2-y1)-(y0-y3)*(x2-x1))
  if (b > a) then a = b
  dim as integer n = sqr(a)/8
  if (n > 1) then
    n = (n+1) and &HFFE
    'if (n > 100) then n = 100 ' make huge curves not hang forever
    dim as single e = 1.0/n,ee=e*e,eee=ee*e
    ' calculate the coefficients of 3rd order equation:
    dim as single xa =   (x3-3*x2+3*x1-x0)
    dim as single xb = 3*(x2-2*x1+x0)
    dim as single xc = 3*(x1-x0)
    ' calculate the forward differences:
    dim as single dx1 = ((xa*e+xb)*e+xc)*e
    dim as single dx3 = 6*xa*eee
    dim as single dx2 = dx3 + 2*xb*ee
    ' calculate the coefficients of 3rd order equation:
    dim as single ya = (y3-3*y2+3*y1-y0)
    dim as single yb = 3*(y2-2*y1+y0)
    dim as single yc = 3*(y1-y0)
    ' calculate the forward differences:
    dim as single dy1 = ((ya*e+yb)*e+yc)*e
    dim as single dy3 = 6*ya*eee
    dim as single dy2 = dy3 + 2*yb*ee
    
    for m as integer=0 to n
      xa=x0:xb=y0
      x0 += dx1 : dx1 += dx2 : dx2 += dx3
      y0 += dy1 : dy1 += dy2 : dy2 += dy3
      this.DrawLine(xa,xb,x0,y0)
    next
  end if
end sub

sub RenderContext.DrawArc(byval cx as integer, byval cy as integer, byval radius as integer, byval angleStart as integer, byval angleEnd as integer)
  if radius<0 then radius=-radius
  if radius<1 then return
  if angleStart=angleEnd then return
  if angleStart>angleEnd then swap angleStart,angleEnd
  dim as integer angle = angleEnd-angleStart
  if angle>358 then this.DrawCircle(cx,cy,radius) : return 
  dim as single  w  = angleStart*DEG2RAD
  dim as integer iX = cx+cos(w)*radius
  dim as integer iY = cy+sin(w)*radius
  for i as single = angleStart+1 to angleEnd
    w=i*DEG2RAD
    dim as integer iX2=cx+cos(w)*radius
    dim as integer iY2=cy+sin(w)*radius
    this.DrawLine(iX,iY,iX2,iY2)
    iX=iX2:iY=iY2
  next
end sub

sub RenderContext.DrawPie(byval cx as integer, byval cy as integer, byval radius as integer, byval angleStart as integer, byval angleEnd as integer)
  if radius<0 then radius=-radius
  if radius<1 then return
  if angleStart=angleEnd then return
  if angleStart>angleEnd then swap angleStart,angleEnd
  dim as integer angle = angleEnd-angleStart
  if angle>358 then this.DrawCircle(cx,cy,radius) : return
  dim as integer steps=iif(angle and 3=0,4,iif(angle and 1=0,2,1))
  dim as single  w  = angleStart*DEG2RAD
  dim as integer iX = cx+cos(w)*radius
  dim as integer iY = cy+sin(w)*radius
  this.DrawLine(cx,cy,iX,iY)
  for i as single = angleStart+steps to angleEnd step steps
    w=i*DEG2RAD
    dim as integer iX2=cx+cos(w)*radius
    dim as integer iY2=cy+sin(w)*radius
    this.DrawLine(iX,iY,iX2,iY2)
    iX=iX2:iY=iY2
  next
  this.DrawLine(iX,iY,cx,cy)
end sub

sub RenderContext.FillPie(byval cx as integer, byval cy as integer, byval radius as integer, byval angleStart as integer, byval angleEnd as integer, byval Boarder as boolean)
  if angleStart=angleEnd then return
  if radius<0 then radius=-radius
  if radius<1 then return
  if angleStart>angleEnd then swap angleStart,angleEnd
  dim as integer angle = angleEnd-angleStart
  if angle>358 then this.FillCircle(cx,cy,radius,boarder) : return 
  dim as integer steps=iif(angle and 3=0,4,iif(angle and 1=0,2,1))
  dim as single  w  = angleStart*DEG2RAD
  dim as integer iX = cx+cos(w)*radius
  dim as integer iY = cy+sin(w)*radius
  for i as single = anglestart+steps to angleEnd step steps
    w=i*DEG2RAD
    dim as integer iX2=cx+cos(w)*radius
    dim as integer iY2=cy+sin(w)*radius
    this.FillTriangle(cx,cy,iX,iY,iX2,iY2)
    iX=iX2:iY=iY2
  next
  if Boarder then
    swap this.DrawingColor.u32,this.BoarderColor.u32
    DrawPie(cx,cy,radius,angleStart,angleEnd)
    swap this.DrawingColor.u32,this.BoarderColor.u32
  end if
end sub


' ##################
' # vector drawing #
' ##################
sub RenderContext.PushMatrix
  if (this.MatrixStackIndex=this.MATRIXSTACK_SIZE) then
    WindowTitle "error: PushMatrix stack overflow." : beep
  else
    this.MatrixStack(this.MatrixStackIndex) = this.CurrentMatrix
    this.MatrixStackIndex+=1
  end if
end sub

sub RenderContext.PopMatrix
  if (this.MatrixStackIndex=0) then
    WindowTitle "error: PopMatrix stack underflow." : beep
  else 
    this.MatrixStackIndex-=1
    this.CurrentMatrix = this.MatrixStack(this.MatrixStackIndex)
  end if
end sub

sub RenderContext.Identiy
  this.CurrentMatrix.a = 1 : this.CurrentMatrix.b = 0
  this.CurrentMatrix.c = 0 : this.CurrentMatrix.d = 1
  this.CurrentMatrix.x = 0 : this.CurrentMatrix.y = 0
end sub

sub RenderContext.Rotate(byval degree as single)
  if (degree) then
    dim as single s=any, c=any
    if (degree=0) then
      s = 0 : c = 1
    elseif (degree=90) then
      s = 1 : c = 0
    elseif (degree=180) then
      s = 0 : c =-1
    elseif (degree=270) or (degree=-90) then
      s =-1 : c = 0
    else
      degree *=DEG2RAD
      s = sin(degree)
      c = cos(degree)
    end if
    this.MultMatrix(c,-s, _
                    s, c, _
                    0, 0)
  end if
end sub

sub RenderContext.Scale(byval s as single)
  this.Scale(s,s)
end sub
sub RenderContext.Scale(byval sx as single,byval sy as single)
  this.MultMatrix(sx,0, 0,sy, 0,0)
end sub
sub RenderContext.Translate(byval tx as single,byval ty as single)
  this.MultMatrix(1,0, 0,1, tx,ty)
end sub

function RenderContext.TransformX(byval x as single, byval y as single) as single
  return x*this.CurrentMatrix.a + y*this.CurrentMatrix.c + this.CurrentMatrix.x
end function
function RenderContext.TransformY(byval x as single, byval y as single) as single
  return x*this.CurrentMatrix.b + y*this.CurrentMatrix.d + this.CurrentMatrix.y
end function


function RenderContext.TransformDX(byval x as single, byval y as single) as single
  return x*this.CurrentMatrix.a + y*this.CurrentMatrix.c
end function
function RenderContext.TransformDY(byval x as single, byval y as single) as single
  return x*this.CurrentMatrix.b + y*this.CurrentMatrix.d
end function

sub RenderContext.DrawPoints()
  if this.points=0 then return
  if this.nPoints<1 then return
  dim as integer i
  while i<this.nPoints
    this.DrawPixel(this.points[i].x,this.points[i].y)
    i+=1
  wend
end sub

sub RenderContext.DrawLines()
  if this.points=0 then return
  if this.nPoints<2 then DrawPoints() : return
  dim as integer i
  while i<this.nPoints-1
    this.DrawLine(this.points[i].x,this.points[i].y, this.points[i+1].x,this.points[i+1].y)
    i+=1
  wend
end sub

sub RenderContext.FillPolygon()
  if this.points=0  then return
  if this.nPoints<3 then return
  select case as const this.current.iBytes
  case 1 : this.FillPolygon8
  case 2 : this.FillPolygon16
  case 4 : this.FillPolygon32
  end select
end sub

sub RenderContext.Gap
  while (this.nPoints > this.gaps+2) andalso _
        (this.points[this.nPoints-1].x = this.points[this.gaps].x) andalso _
        (this.points[this.nPoints-1].y = this.points[this.gaps].y)
    this.nPoints-=1
  wend
  if (this.nPoints > (this.gaps+2)) then
    this.AddTransformedVertex(this.points[this.gaps].x, this.points[this.gaps].y)
    this.gaps = this.nPoints
  else
    this.nPoints = this.gaps
  end if
end sub

sub Rendercontext.AddVertex(byval x as single, byval y as single)
  dim as integer iX = x*this.CurrentMatrix.a + y*this.CurrentMatrix.c + this.CurrentMatrix.x
  dim as integer iY = x*this.CurrentMatrix.b + y*this.CurrentMatrix.d + this.CurrentMatrix.y
  this.AddTransformedVertex(iX,iY)
end sub

sub Rendercontext.AddTransformedVertex(byval iX as integer, byval iY as integer)
  if this.Points =  0 then return
  if this.nPoints=this.nMaxPoints then 
    this.nMaxPoints shl = 1
    this.Points=reallocate(this.Points,sizeof(Point2D)*this.nMaxPoints)
  end if
  this.points[this.nPoints].x=iX
  this.points[this.nPoints].y=iY
  this.nPoints+=1
end sub

sub RenderContext.BeginPoints
  this.vMode=vPOINTS : this.nPoints=0
end sub
sub RenderContext.EndPoints
  if (this.vMode and this.vPOINTS)=0 then return ' missing BeginPoints
  if this.nPoints>0 then this.DrawPoints()
end sub

sub RenderContext.BeginLines
  this.vMode=vLINES : this.nPoints=0
end sub
sub RenderContext.EndLines
  if (this.vMode and this.vLINES)=0 then return ' missing BeginLines
  if this.nPoints<1 then return

  if this.nPoints=1 then
    this.DrawPoints()
  else
    this.DrawLines()
  end if
end sub

sub RenderContext.BeginLineLoop
  this.vMode=vLINELOOP : this.nPoints=0
end sub

sub RenderContext.EndLineLoop
  if (this.vMode and this.vLINELOOP)=0 then return ' missing BeginLineLoop
  while (this.nPoints>2) andalso (this.points[this.nPoints-1].x = this.points[0].x) _
                         andalso (this.points[this.nPoints-1].y = this.points[0].y)
    this.nPoints-=1
  wend
  if (this.nPoints>2) then this.AddTransformedVertex(this.points[0].x, this.points[0].y)
  this.DrawLines()
end sub

sub RenderContext.BeginPolygon
  this.vMode=vPOLYGON : this.nPoints=0
end sub
sub RenderContext.EndPolygon
  if (this.vMode and this.vPOLYGON)=0 then return ' missing BeginPolygon
  while (this.nPoints>2) andalso (this.points[this.nPoints-1].x = this.points[0].x) _
                         andalso (this.points[this.nPoints-1].y = this.points[0].y)
    this.nPoints-=1
  wend
  if (this.nPoints<3) then DrawLines() : return
  this.DrawLines()
end sub

sub RenderContext.BeginComplexPolygon
  this.vMode=vPOLYGON or vCOMPLEX : this.nPoints=0 : this.Gaps = 0
end sub
sub RenderContext.EndComplexPolygon
  if (this.vMode and this.vPOLYGON)=0 then return
  if (this.vMode and this.vCOMPLEX)=0 then return ' missing BeginComplexPolygon
  this.gap()
  if (this.nPoints < 3) then DrawLines() : return
  this.FillPolygon()
end sub


sub RenderContext.AddCurve(byval x0 as single, byval y0 as single, _
                           byval x1 as single, byval y1 as single, _
                           byval x2 as single, byval y2 as single, _
                           byval x3 as single, byval y3 as single)

  dim as single tx0 = this.TransformX(x0,y0)
  dim as single ty0 = this.TransformY(x0,y0)

  dim as single tx1  = this.TransformX(x1,y1)
  dim as single ty1  = this.TransformY(x1,y1)
  
  dim as single tx2  = this.TransformX(x2,y2)
  dim as single ty2  = this.TransformY(x2,y2)

  dim as single tx3  = this.TransformX(x3,y3)
  dim as single ty3  = this.TransformY(x3,y3)

  ' find area
  dim as single a = abs((tx0-tx2)*(ty3-ty1)-(ty0-ty2)*(tx3-tx1))
  dim as single b = abs((tx0-tx3)*(ty2-ty1)-(ty0-ty3)*(tx2-tx1))
  if (b > a) then a = b

  ' use that to guess at the number of segments:
  dim as integer n = sqr(a)/8
  if (n > 1) then
    n = (n+1) and &HFFE
    dim as single e = 1.0/n,ee=e*e,eee=ee*e
    ' calculate the x coefficients of 3rd order equation:
    dim as single xa =   (tx3-3*tx2+3*tx1-tx0)
    dim as single xb = 3*(tx2-2*tx1+tx0)
    dim as single xc = 3*(tx1-tx0)
    ' calculate the x forward differences:
    dim as single dx1 = ((xa*e+xb)*e+xc)*e
    dim as single dx3 = 6*xa*eee
    dim as single dx2 = dx3 + 2*xb*ee
    ' calculate the y coefficients of 3rd order equation:
    dim as single ya = (ty3-3*ty2+3*ty1-ty0)
    dim as single yb = 3*(ty2-2*ty1+ty0)
    dim as single yc = 3*(ty1-ty0)
    ' calculate the y forward differences:
    dim as single dy1 = ((ya*e+yb)*e+yc)*e
    dim as single dy3 = 6*ya*eee
    dim as single dy2 = dy3 + 2*yb*ee
    this.AddTransformedVertex(tx0,ty0)
    for m as integer=0 to n
      tx0 += dx1 : dx1 += dx2 : dx2 += dx3
      ty0 += dy1 : dy1 += dy2 : dy2 += dy3
      this.AddTransformedVertex(tx0,ty0)
    next
  end if
end sub


sub RenderContext.AddPie(byval cx as single, byval cy as single, byval radius as single, byval angleStart as single, byval angleEnd as single)
  if radius<0 then radius=-radius
  if radius<0.01 then return
  if angleStart=angleEnd then return
  if angleStart>angleEnd then swap angleStart,angleEnd
  dim as boolean IsClosed
  if (angleEnd-angleStart)>360 then 
     angleEnd=angleStart+360
     IsClosed=true
  end if
  if IsClosed=false then this.AddVertex(cx,cy)
  for i as single = angleStart to angleEnd
     dim as single w=i*DEG2RAD
     dim as single x=cx+cos(w)*radius
     dim as single y=cy+sin(w)*radius
     this.AddVertex(x,y)
  next
  if IsClosed=false then this.AddVertex(cx,cy)
end sub

sub RenderContext.AddArc(byval cx as single, byval cy as single, byval radius as single, byval angleStart as single, byval angleEnd as single)
  if radius<0 then radius=-radius
  if radius<0.01 then return
  if angleStart=angleEnd then return
  if angleStart>angleEnd then swap angleStart,angleEnd
  if (angleEnd-angleStart)>360 then angleEnd=angleStart+360
  for i as single = angleStart to angleEnd
     dim as single w=i*DEG2RAD
     dim as single x=cx+cos(w)*radius
     dim as single y=cy+sin(w)*radius
     this.AddVertex(x,y)
  next
end sub

' #########
' # tools #
' #########
function RenderContext.AngleDegree(byval x as integer, byval y as integer) as integer
  if (x=0) andalso (y=0) then 
    return 0
  elseif (x=0) then
    return iif(y>0,90,270)
  elseif (y=0) andalso (x<0) then
    return 180
  else
    dim as integer a = atn(y/x)*RAD2DEG
    return iif(x<0,a+180,iif(a<0,a+360,a))
  end if
end function

function RenderContext.AngleDegree(byval x1 as integer, byval y1 as integer, _
                                   byval x2 as integer, byval y2 as integer) as integer
  return AngleDegree(x2-x1,y2-y1)
end function


#define CRS(A,B,C,D,T) ((((-.5*(A)+1.5*(B)+-1.5*(C)+.5*(D))*(T)+(A)+-2.5*(B)+2.0*(C)+-.5*(D))*(T)+(-.5*(A)+.5*(C)))*(T)+(B))

#endif ' __FBGFX_ADDON_BI__

